#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <pthread.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "types.h"
#include "uts.h"
#include "mem.h"
#include "fdset.h"
#include "ana_io.h"
//#include "fits_read.h"
#include "fillpix.h"
#include "momfbd_read.h"
#include "cmdcfg.h"

void inv4x4(float64_t *mat, float64_t *dst)
{
  float64_t tmp[12]; // temp array for pairs
  float64_t src[16]; // array of transpose source matrix
  float64_t det;     // determinant
  
// transpose matrix
  for(int i=0;i<4;i++){
    src[i   ]=mat[i*4  ];
    src[i+ 4]=mat[i*4+1];
    src[i+ 8]=mat[i*4+2];
    src[i+12]=mat[i*4+3];
  }
  
// calculate pairs for first 8 elements (cofactors)
  tmp[ 0]=src[10]*src[15];
  tmp[ 1]=src[11]*src[14];
  tmp[ 2]=src[ 9]*src[15];
  tmp[ 3]=src[11]*src[13];
  tmp[ 4]=src[ 9]*src[14];
  tmp[ 5]=src[10]*src[13];
  tmp[ 6]=src[ 8]*src[15];
  tmp[ 7]=src[11]*src[12];
  tmp[ 8]=src[ 8]*src[14];
  tmp[ 9]=src[10]*src[12];
  tmp[10]=src[ 8]*src[13];
  tmp[11]=src[ 9]*src[12];
// calculate first 8 elements (cofactors)
  dst[0] =tmp[0]*src[5]+tmp[3]*src[6]+tmp[ 4]*src[7];
  dst[0]-=tmp[1]*src[5]+tmp[2]*src[6]+tmp[ 5]*src[7];
  dst[1] =tmp[1]*src[4]+tmp[6]*src[6]+tmp[ 9]*src[7];
  dst[1]-=tmp[0]*src[4]+tmp[7]*src[6]+tmp[ 8]*src[7];
  dst[2] =tmp[2]*src[4]+tmp[7]*src[5]+tmp[10]*src[7];
  dst[2]-=tmp[3]*src[4]+tmp[6]*src[5]+tmp[11]*src[7];
  dst[3] =tmp[5]*src[4]+tmp[8]*src[5]+tmp[11]*src[6];
  dst[3]-=tmp[4]*src[4]+tmp[9]*src[5]+tmp[10]*src[6];
  dst[4] =tmp[1]*src[1]+tmp[2]*src[2]+tmp[ 5]*src[3];
  dst[4]-=tmp[0]*src[1]+tmp[3]*src[2]+tmp[ 4]*src[3];
  dst[5] =tmp[0]*src[0]+tmp[7]*src[2]+tmp[ 8]*src[3];
  dst[5]-=tmp[1]*src[0]+tmp[6]*src[2]+tmp[ 9]*src[3];
  dst[6] =tmp[3]*src[0]+tmp[6]*src[1]+tmp[11]*src[3];
  dst[6]-=tmp[2]*src[0]+tmp[7]*src[1]+tmp[10]*src[3];
  dst[7] =tmp[4]*src[0]+tmp[9]*src[1]+tmp[10]*src[2];
  dst[7]-=tmp[5]*src[0]+tmp[8]*src[1]+tmp[11]*src[2];
// calculate pairs for second 8 elements (cofactors)
  tmp[ 0]=src[2]*src[7];
  tmp[ 1]=src[3]*src[6];
  tmp[ 2]=src[1]*src[7];
  tmp[ 3]=src[3]*src[5];
  tmp[ 4]=src[1]*src[6];
  tmp[ 5]=src[2]*src[5];
  tmp[ 6]=src[0]*src[7];
  tmp[ 7]=src[3]*src[4];
  tmp[ 8]=src[0]*src[6];
  tmp[ 9]=src[2]*src[4];
  tmp[10]=src[0]*src[5];
  tmp[11]=src[1]*src[4];
  /* calculate second 8 elements (cofactors) */
  dst[ 8] =tmp[ 0]*src[13]+tmp[ 3]*src[14]+tmp[ 4]*src[15];
  dst[ 8]-=tmp[ 1]*src[13]+tmp[ 2]*src[14]+tmp[ 5]*src[15];
  dst[ 9] =tmp[ 1]*src[12]+tmp[ 6]*src[14]+tmp[ 9]*src[15];
  dst[ 9]-=tmp[ 0]*src[12]+tmp[ 7]*src[14]+tmp[ 8]*src[15];
  dst[10] =tmp[ 2]*src[12]+tmp[ 7]*src[13]+tmp[10]*src[15];
  dst[10]-=tmp[ 3]*src[12]+tmp[ 6]*src[13]+tmp[11]*src[15];
  dst[11] =tmp[ 5]*src[12]+tmp[ 8]*src[13]+tmp[11]*src[14];
  dst[11]-=tmp[ 4]*src[12]+tmp[ 9]*src[13]+tmp[10]*src[14];
  dst[12] =tmp[ 2]*src[10]+tmp[ 5]*src[11]+tmp[ 1]*src[ 9];
  dst[12]-=tmp[ 4]*src[11]+tmp[ 0]*src[ 9]+tmp[ 3]*src[10];
  dst[13] =tmp[ 8]*src[11]+tmp[ 0]*src[ 8]+tmp[ 7]*src[10];
  dst[13]-=tmp[ 6]*src[10]+tmp[ 9]*src[11]+tmp[ 1]*src[ 8];
  dst[14] =tmp[ 6]*src[ 9]+tmp[11]*src[11]+tmp[ 3]*src[ 8];
  dst[14]-=tmp[10]*src[11]+tmp[ 2]*src[ 8]+tmp[ 7]*src[ 9];
  dst[15] =tmp[10]*src[10]+tmp[ 4]*src[ 8]+tmp[ 9]*src[ 9];
  dst[15]-=tmp[ 8]*src[ 9]+tmp[11]*src[10]+tmp[ 5]*src[ 8];
// calculate determinant 
  det=src[0]*dst[0]+src[1]*dst[1]+src[2]*dst[2]+src[3]*dst[3];
// calculate matrix inverse 
  for(int j=0;j<16;j++) dst[j]/=det;
}

#define PI               3.1415

double pythag(double a, double b)
{
  double absa,absb;
  absa=fabs(a);
  absb=fabs(b);
  if (absa > absb) return absa*sqrt(1.0+sqr(absb/absa));
  else return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+sqr(absa/absb)));
}

void svdcmp(double **a,int m,int n,double *w,double **v)
{
  int flag,i,its,j,jj,k,l,nm;
  double anorm,c,f,g,h,s,scale,x,y,z,*rv1;

  rv1=d1dim(1,n);
  g=scale=anorm=0.0;
  for(i=1;i<=n;i++){
    l=i+1;
    rv1[i]=scale*g;
    g=s=scale=0.0;
    if(i<=m){
      for(k=i;k<=m;k++) scale+=fabs(a[k][i]);
      if(scale){
        for(k=i;k<=m;k++){
          a[k][i]/=scale;
          s+=a[k][i]*a[k][i];
        }
        f=a[i][i];
        g=-sign(sqrt(s),f);
        h=f*g-s;
        a[i][i]=f-g;
        for(j=l;j<=n;j++){
          for(s=0.0,k=i;k<=m;k++) s+=a[k][i]*a[k][j];
          f=s/h;
          for(k=i;k<=m;k++) a[k][j]+=f*a[k][i];
        }
        for(k=i;k<=m;k++) a[k][i]*=scale;
      }
    }
    w[i]=scale*g;
    g=s=scale=0.0;
    if ((i<=m)&&(i!=n)) {
      for(k=l;k<=n;k++) scale+=fabs(a[i][k]);
      if(scale){
        for (k=l;k<=n;k++){
          a[i][k]/=scale;
          s+=a[i][k]*a[i][k];
        }
        f=a[i][l];
        g=-sign(sqrt(s),f);
        h=f*g-s;
        a[i][l]=f-g;
        for(k=l;k<=n;k++) rv1[k]=a[i][k]/h;
        for(j=l;j<=m;j++) {
          for(s=0.0,k=l;k<=n;k++) s+=a[j][k]*a[i][k];
          for(k=l;k<=n;k++) a[j][k]+=s*rv1[k];
        }
        for(k=l;k<=n;k++) a[i][k]*=scale;
      }
    }
    anorm=fmax(anorm,(fabs(w[i])+fabs(rv1[i])));
  }
  for (i=n;i>=1;i--) {
    if (i < n) {
      if (g) {
        for (j=l;j<=n;j++) v[j][i]=(a[i][j]/a[i][l])/g;
        for (j=l;j<=n;j++) {
          for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
          for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
        }
      }
      for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
    }
    v[i][i]=1.0;
    g=rv1[i];
    l=i;
  }
  for (i=min(m,n);i>=1;i--) {
    l=i+1;
    g=w[i];
    for (j=l;j<=n;j++) a[i][j]=0.0;
    if (g) {
      g=1.0/g;
      for (j=l;j<=n;j++) {
        for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
        f=(s/a[i][i])*g;
        for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
      }
      for (j=i;j<=m;j++) a[j][i] *= g;
    } else for (j=i;j<=m;j++) a[j][i]=0.0;
    ++a[i][i];
  }
  for (k=n;k>=1;k--) {
    for (its=1;its<=30;its++) {
      flag=1;
      for (l=k;l>=1;l--) {
        nm=l-1;
        if ((double)(fabs(rv1[l])+anorm) == anorm) {
          flag=0;
          break;
        }
        if ((double)(fabs(w[nm])+anorm) == anorm) break;
      }
      if (flag) {
        c=0.0;
        s=1.0;
        for (i=l;i<=k;i++) {
          f=s*rv1[i];
          rv1[i]=c*rv1[i];
          if ((double)(fabs(f)+anorm) == anorm) break;
          g=w[i];
          h=pythag(f,g);
          w[i]=h;
          h=1.0/h;
          c=g*h;
          s = -f*h;
          for (j=1;j<=m;j++) {
            y=a[j][nm];
            z=a[j][i];
            a[j][nm]=y*c+z*s;
            a[j][i]=z*c-y*s;
          }
        }
      }
      z=w[k];
      if (l == k) {
        if (z < 0.0) {
          w[k] = -z;
          for (j=1;j<=n;j++) v[j][k] = -v[j][k];
        }
        break;
      }
      if(its==30) fprintf(stderr,"no convergence in 30 svdcmp iterations\n");
      x=w[l];
      nm=k-1;
      y=w[nm];
      g=rv1[nm];
      h=rv1[k];
      f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
      g=pythag(f,1.0);
      f=((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x;
      c=s=1.0;
      for (j=l;j<=nm;j++) {
        i=j+1;
        g=rv1[i];
        y=w[i];
        h=s*g;
        g=c*g;
        z=pythag(f,h);
        rv1[j]=z;
        c=f/z;
        s=h/z;
        f=x*c+g*s;
        g = g*c-x*s;
        h=y*s;
        y *= c;
        for (jj=1;jj<=n;jj++) {
          x=v[jj][j];
          z=v[jj][i];
          v[jj][j]=x*c+z*s;
          v[jj][i]=z*c-x*s;
        }
        z=pythag(f,h);
        w[j]=z;
        if (z) {
          z=1.0/z;
          c=f*z;
          s=h*z;
        }
        f=c*g+s*y;
        x=c*y-s*g;
        for (jj=1;jj<=m;jj++) {
          y=a[jj][j];
          z=a[jj][i];
          a[jj][j]=y*c+z*s;
          a[jj][i]=z*c-y*s;
        }
      }
      rv1[l]=0.0;
      rv1[k]=f;
      w[k]=x;
    }
  }
  del_d1dim(rv1,1,n);
}

void svbksb(double **u, double w[], double **v, int m, int n, double b[], double x[])
{
  int jj,j,i;
  double s,*tmp;

  tmp=d1dim(1,n);
  for (j=1;j<=n;j++) {
    s=0.0;
    if (w[j]) {
      for (i=1;i<=m;i++) s += u[i][j]*b[i];
      s /= w[j];
    }
    tmp[j]=s;
  }
  for (j=1;j<=n;j++) {
    s=0.0;
    for (jj=1;jj<=n;jj++) s += v[j][jj]*tmp[jj];
    x[j]=s;
  }
  del_d1dim(tmp,1,n);
}

void swap(int *a,int *b)
{
  int c=*a;
  *a=*b;
  *b=c;
}

void svd(double **AA,int n,double eps,FILE *f)
{
  double **U=d2dim(1,n,1,n),**V=d2dim(1,n,1,n);
  double *x=d1dim(1,n),*y=d1dim(1,n),*z=d1dim(1,n),*w=d1dim(1,n),*wi=d1dim(1,n);
  int *ls=i1dim(1,n);
  int i,j,k;
  double wicut;
/* sort diagonal elements, so that larger values are located in the top left hand corner of A resp. U */
  for(i=1;i<=n;++i) ls[i]=i;
#ifndef NOSORT
  for(i=1;i<n;++i) for(j=i+1;j<=n;++j) if(AA[ls[j]][ls[j]]>AA[ls[i]][ls[i]]) swap(ls+i,ls+j);
#endif
  for(i=1;i<=n;++i){
    for(j=1;j<=n;++j) U[i][j]=AA[ls[i]][ls[j]];
    x[i]=0.0;
  }
/* SVD u (=a), output in u,w,v */
  svdcmp(U,n,n,w,V);
/* find maximum singular value in each parameter subspace pay attention to the new (sorted) storage of u(i,j) */
  memset(z+1,0,n*sizeof(double));
//
  wicut=0.0;
  for(i=1;i<=n;++i) wicut=fmax(wicut,w[i]);
  wicut*=eps;
  for(i=1;i<=n;++i) if(w[i]>wicut) z[i]=w[i];
//
  if(f){
    fprintf(f,"singular values:\n");
    for(int l=1;l<=n;++l) fprintf(f,"%d: %E -> %E\n",l,w[l],z[l]);
  }
//
  for(i=1;i<=n;++i) w[i]=z[i]; /* replace singular values by modified ones */
  svbksb(U,w,V,n,n,x,y);       /* backsubstitute */ 
  for(i=1;i<=n;++i){
    if(w[i]>0.0) w[i]=1.0/w[i];
  }
/* covar matrix = Inv(A) = [V][1/w][Ut] 
   Covar matrix for error estimates
   See Num Recipes (Press et al.,1992: 15.6.10), p694
   CORRECTION: 1/w_i instead of 1/w_i^2 !!!*/
  for(i=1;i<=n;++i)
    for(j=1;j<=n;++j){
      double sum=0.0;
      for(k=1;k<=n;++k) sum+=V[i][k]*V[j][k]*w[k];
      AA[ls[i]][ls[j]]=sum;
    }
/* cleanup */
  del_i1dim(ls,1,n);
  del_d1dim(wi,1,n);
  del_d1dim(w,1,n);
  del_d1dim(z,1,n);
  del_d1dim(y,1,n);
  del_d1dim(x,1,n);
  del_d2dim(V,1,n,1,n);
  del_d2dim(U,1,n,1,n);
}

void mgsolve(float64_t ****A,float64_t **rhs,float64_t **xx,int npts)
{
  switch(npts){
    case(1):{
      xx[1][1]=rhs[1][1]/A[1][1][1][1];
      break;
    }
    case(2):{
      float64_t ****Ai=f64t4dim(1,npts,1,npts,1,npts,1,npts);
      inv4x4(A[1][1][1]+1,Ai[1][1][1]+1);
      xx[1][1]=rhs[1][1]*Ai[1][1][1][1]+rhs[1][2]*Ai[1][1][1][2]+rhs[2][1]*Ai[1][1][2][1]+rhs[2][2]*Ai[1][1][2][2];
      xx[1][2]=rhs[1][1]*Ai[1][2][1][1]+rhs[1][2]*Ai[1][2][1][2]+rhs[2][1]*Ai[1][2][2][1]+rhs[2][2]*Ai[1][2][2][2];
      xx[2][1]=rhs[1][1]*Ai[2][1][1][1]+rhs[1][2]*Ai[2][1][1][2]+rhs[2][1]*Ai[2][1][2][1]+rhs[2][2]*Ai[2][1][2][2];
      xx[2][2]=rhs[1][1]*Ai[2][2][1][1]+rhs[1][2]*Ai[2][2][1][2]+rhs[2][1]*Ai[2][2][2][1]+rhs[2][2]*Ai[2][2][2][2];
      break;
    }
    case(3):{
      float64_t **Ai=f64t2dim(1,npts*npts,1,npts*npts);
      for(int x=1;x<=npts;++x) for(int y=1;y<=npts;++y){
        float64_t *p=Ai[1+(x-1)*npts+(y-1)];
        for(int u=1;u<=npts;++u) for(int v=1;v<=npts;++v) p[1+(u-1)*npts+(v-1)]=A[x][y][u][v];
      }
      svd(Ai,npts*npts,1E-3,0);
      for(int x=1;x<=npts;++x) for(int y=1;y<=npts;++y){
        float64_t *p=Ai[1+(x-1)*npts+(y-1)];
        xx[x][y]=0.0;
        for(int u=1;u<=npts;++u) for(int v=1;v<=npts;++v) xx[x][y]+=p[1+(u-1)*npts+(v-1)]*rhs[u][v];
      }
      break;
    }
    default:{
      int np=npts/2;
      float64_t **dfct=f64t2dim(1,npts,1,npts);
      float64_t ****Ac=f64t4dim(1,np,1,np,1,np,1,np);
      float64_t **dc=f64t2dim(1,np,1,np),**xc=f64t2dim(1,np,1,np);
      memset(xc[1]+1,0,np*np*sizeof(float64_t));
  
      for(int x=1;x<=npts;++x) for(int y=1;y<=npts;++y){ // defect = rhs - A * x;
        dfct[x][y]=rhs[x][y];
        for(int u=1;u<=npts;++u) for(int v=1;v<=npts;++v) dfct[x][y]-=A[x][y][u][v]*xx[u][v];
      }
      for(int x=1;x<=np;++x) for(int y=1;y<=np;++y){
        dc[x][y]=0.0;
        for(int dx=-1;dx<=0;++dx) for(int dy=-1;dy<=0;++dy) dc[x][y]+=dfct[2*x+dx][2*y+dy];
        for(int u=1;u<=np;++u) for(int v=1;v<=np;++v){
          Ac[x][y][u][v]=0.0;
          for(int dx=-1;dx<=0;++dx) for(int dy=-1;dy<=0;++dy)
            for(int du=-1;du<=0;++du) for(int dv=-1;dv<=0;++dv) Ac[x][y][u][v]+=A[2*x+dx][2*y+dy][2*u+du][2*v+dv];
        }
      }
      mgsolve(Ac,dc,xc,np);
    }
  }
}

fp_t **fit_surface(fp_t **a,fp_t *X,fp_t *Y,int nx,int ny,int dx,int dy)
{
  fp_t *xx=new fp_t [2*dx+2];
  fp_t **x=ft2dim(1,nx,0,dx);
  memset(xx,0,(2*dx+2)*sizeof(fp_t));
  for(int i=1;i<=nx;++i){
    xx[0]+=(x[i][0]=1.0);
    for(int j=1;j<=dx;++j) xx[j]+=(x[i][j]=x[i][j-1]*X[i]);
    for(int j=dx+1;j<=2*dx;++j) xx[j]+=x[i][j-dx]*x[i][dx];
  }
  fp_t *yy=new fp_t [2*dy+2];
  fp_t **y=ft2dim(1,ny,0,dy);
  memset(yy,0,(2*dy+2)*sizeof(fp_t));
  for(int i=1;i<=ny;++i){
    yy[0]+=(y[i][0]=1.0);
    for(int j=1;j<=dy;++j) yy[j]+=(y[i][j]=y[i][j-1]*Y[i]);
    for(int j=dy+1;j<=2*dy;++j) yy[j]+=y[i][j-dy]*y[i][dy];
  }
//
  int N=(dx+1)*(dy+1);
  fp_t **alpha=ft2dim(1,N,1,N);
  fp_t *beta=new fp_t [N]-1;
  fp_t *row=new fp_t [ny]-1;
//  
  int inc1=1;
  for(int i=0;i<=dx;++i)
    for(int j=0;j<=dy;++j){
      if(!j)
        for(int iy=1;iy<=ny;++iy){
          row[iy]=0.0;
          for(int ix=1;ix<=nx;++ix) row[iy]+=x[ix][i]*a[ix][iy];
        }
      beta[inc1]=0.0;
      for(int iy=1;iy<=ny;++iy) beta[inc1]+=row[iy]*y[iy][j];
      int inc2=1;
      for(int k=0;k<=dx;++k)
        for(int l=0;l<=dy;++l)
 	  alpha[inc1][inc2++]=xx[i+k]*yy[j+l];	  // Matrix to be LU decomposed.
      ++inc1;
    }
  del_ft2dim(x,1,nx,0,dx);
  del_ft2dim(y,1,ny,0,dy);
  delete (row+1);
  delete xx;
  delete yy;
  fp_t det;
  int *idx=new int [N]-1;
  ludcmp(alpha,N,idx,det);                          // Replaces alfa with its LU decompose
  lubksb(alpha,N,idx,beta);                        // Solves the set of linear equations.
  delete (idx+1);
  del_ft2dim(alpha,1,N,1,N);
  return ft2dim(beta+1,0,dx,0,dy);
}

void pm_plane(fp_t **si,fp_t **coefs,int size,fp_t sign)
{
  for(int x=1;x<=size;++x){
    fp_t xx=((fp_t)(x-1)/(fp_t)(size-1))-0.5;
    for(int y=1;y<=size;++y){
      fp_t yy=((fp_t)(y-1)/(fp_t)(size-1))-0.5;
      si[x][y]+=sign*(coefs[0][0]+coefs[1][0]*xx+coefs[0][1]*yy+coefs[1][1]*xx*yy);
    }
  }
}


void norm(float64_t **im,int nx,int ny)
{
  float64_t sum=0.0;
  for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) sum+=im[x][y];
  sum/=((float64_t)nx)*((float64_t)ny);
  for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) im[x][y]/=sum;
  fprintf(stderr,"norm: mean=%f\n",sum);
}

float64_t **unsharp(float64_t **im,int nx,int ny,int dl)
{
  float64_t **usm=f64t2dim(1,nx,1,ny);
  for(int x=1;x<=nx;++x)
    for(int y=1;y<=ny;++y){
      float64_t sum=0.0,nem=0.0;
      for(int i=max(-dl,-x+1);i<=min(dl,nx-x);++i)
        for(int j=max(-dl,-y+1);j<=min(dl,ny-y);++j){
          sum+=im[x+i][y+j];
          nem+=1.0;
        }
      usm[x][y]=im[x][y]-sum/nem;
    }
  return usm;
}


float64_t *correlate(float64_t **im,float64_t **ref,int nx,int ny,int r,int xl,int yl,int xh,int yh)
{
   float64_t IA[9][9]={{0.25,-0.50, 0.25,-0.5, 1.0,-0.5, 0.25,-0.5, 0.25},\
                       {0.25, 0.00,-0.25,-0.5, 0.0, 0.5, 0.25, 0.0,-0.25},\
                       {0.00, 0.50, 0.00, 0.0,-1.0, 0.0, 0.00, 0.5, 0.00},\
                       {0.25,-0.50, 0.25, 0.0, 0.0, 0.0,-0.25, 0.5,-0.25},\
                       {0.25, 0.00,-0.25, 0.0, 0.0, 0.0,-0.25, 0.0, 0.25},\
                       {0.00, 0.50, 0.00, 0.0, 0.0, 0.0, 0.00,-0.5, 0.00},\
                       {0.00, 0.00, 0.00, 0.5,-1.0, 0.5, 0.00, 0.0, 0.00},\
                       {0.00, 0.00, 0.00, 0.5, 0.0,-0.5, 0.00, 0.0, 0.00},\
                       {0.00, 0.00, 0.00, 0.0, 1.0, 0.0, 0.00, 0.0, 0.00}};
  float64_t **cf=f64t2dim(-r,r,-r,r);
  float64_t mv=1.0E32;
  int xm,ym;
  for(int i=-r;i<=r;++i)
    for(int j=-r;j<=r;++j){
      cf[i][j]=0.0;
//      int xl=max(1-i,1+xclip),xh=min(nx-i,nx-xclip);
//      int yl=max(1-j,1+yclip),yh=min(ny-j,ny-yclip);
      for(int x=xl;x<=xh;++x) for(int y=yl;y<=yh;++y) cf[i][j]+=sqr(im[x+i][y+j]-ref[x][y]);
      cf[i][j]/=(float64_t)(xh-xl+1)*(float64_t)(yh-yl+1);
      if(mv>cf[i][j]){
        mv=cf[i][j];
        xm=i;
        ym=j;
      }
//      fprintf(stderr,"cc: %d %d %f\n",i,j,cf[i][j]);
    }
//  fprintf(stderr,"cc: %d %d\n",xm,ym);
  float64_t f[]={cf[xm+1][ym+1],cf[xm+1][ym],cf[xm+1][ym-1],\
                 cf[xm  ][ym+1],cf[xm  ][ym],cf[xm  ][ym-1],\
                 cf[xm-1][ym+1],cf[xm-1][ym],cf[xm-1][ym-1]};
  float64_t cfs[]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
  for(int i=0;i<9;++i) for(int j=0;j<9;++j) cfs[i]+=IA[i][j]*f[j];
//
  float64_t *dl=new float64_t [2];
  float64_t xx=0.0,yy=0.0;
  for(int i=0;i<=20;++i){
// ((cfs[0]*y+cfs[1])*y+cfs[2])*x*x+((cfs[3]*y+cfs[4])*y+cfs[5])*x+((cfs[6]*y+cfs[7])*y+cfs[8])
    float64_t aa=((cfs[0]*yy+cfs[1])*yy+cfs[2]);
    float64_t bb=((cfs[3]*yy+cfs[4])*yy+cfs[5]);
    xx=-bb/(2.0*aa);
// ((cfs[0]*x+cfs[3])*x+cfs[6])*y*y+((cfs[1]*x+cfs[4])*x+cfs[7])*y+((cfs[2]*x+cfs[5])*x+cfs[8])
    aa=((cfs[0]*xx+cfs[3])*xx+cfs[6]);
    bb=((cfs[1]*xx+cfs[4])*xx+cfs[7]);
    yy=-bb/(2.0*aa);
  }
//
  dl[0]=xx+float(xm);
  dl[1]=yy+float(ym);
//  
  del_f64t2dim(cf,-r,r,-r,r);
  return dl;
}

void smooth(fp_t **p,int nx1,int nx2,int n)
{
  fp_t **tmp=ft2dim(1,n,1,nx2),norm=(fp_t)n*n;
  int nh=n/2;
  memset(tmp[1]+1,0,nh*nx2*sizeof(fp_t));
  memcpy(tmp[nh+1]+1,p[1]+1,(nh+1)*nx2*sizeof(fp_t));
  for(int x1=1;x1<=nx1;++x1){
    memset(p[x1]+1,0,nx2*sizeof(fp_t));
    for(int x2=1;x2<=nh;++x2)
      for(int sx1=1;sx1<=n;++sx1) for(int sx2=1;sx2<=x2+nh;++sx2) p[x1][x2]+=tmp[sx1][sx2];
    for(int x2=nh+1;x2<=nx2-nh;++x2)
      for(int sx1=1;sx1<=n;++sx1) for(int sx2=x2-nh;sx2<=x2+nh;++sx2) p[x1][x2]+=tmp[sx1][sx2];
    for(int x2=nx2-nh+1;x2<=nx2;++x2)
      for(int sx1=1;sx1<=n;++sx1) for(int sx2=x2-nh;sx2<=nx2;++sx2) p[x1][x2]+=tmp[sx1][sx2];
    for(int x2=1;x2<=nx2;++x2) p[x1][x2]/=norm;
    memmove(tmp[1]+1,tmp[2]+1,(n-1)*nx2*sizeof(fp_t));
    if(x1<(nx1-nh))
      memcpy(tmp[n]+1,p[x1+nh+1]+1,nx2*sizeof(fp_t));
    else
      memset(tmp[n]+1,0,nx2*sizeof(fp_t));
  }
  del_ft2dim(tmp,1,n,1,nx2);
}

struct lst{
  int x,y;
};

char **append(char **a,const char *b)
{
  int la=0;
  if(a) while(a[la]) ++la;
//
  char **tmp=new char* [la+2];
//
  if(la) memcpy(tmp,a,la*sizeof(char*));
  tmp[la]=strcpy(new char [strlen(b)+1],b);
  tmp[la+1]=0;
//
  return tmp;
}

#define CR_LISTED   0x01
#define CR_SELECTED 0x02

void connected_region(fp_t **p,int nx,int ny,int x_s,int y_s)
{
  byte **mask=b2dim(1,nx,1,ny);
  memset(mask[1]+1,0,nx*ny*sizeof(byte));
  struct lst *list=new struct lst [nx*ny] -1;
  memset(list+1,0,nx*ny*sizeof(struct lst));
  int nmax=1;
  list[nmax].x=x_s;
  list[nmax].y=y_s;
  mask[x_s][y_s]|=CR_LISTED|CR_SELECTED;
  for(int m=1;m<=nmax;++m)
    if(p[list[m].x][list[m].y]>0.0){ // pixel is good: check the neighbours
      mask[list[m].x][list[m].y]|=CR_SELECTED;
      for(int i=-1;i<=1;++i)
        if(((list[m].x+i)>=1)&&((list[m].x+i)<=nx))
          for(int j=-1;j<=1;++j)
            if(((list[m].y+j)>=1)&&((list[m].y+j)<=ny))
	      if(!(mask[list[m].x+i][list[m].y+j]&CR_LISTED)){ // pixel has not been listed yet
                ++nmax;                                        // add element to list
                list[nmax].x=list[m].x+i;                      // record pixel coordinates
                list[nmax].y=list[m].y+j;
	        mask[list[m].x+i][list[m].y+j]|=CR_LISTED;     // avoid picking the same pixel twice
              }
    }
  for(int x=1;x<=nx;++x)
    for(int y=1;y<=ny;++y)
      if(!(mask[x][y]&CR_SELECTED)) p[x][y]=0.0;
  del_b2dim(mask,1,nx,1,ny);
  delete[] (list+1);
}

#undef CR_LISTED
#undef CR_SELECTED

void ScharmerFilter(complex_t **obj_f,fp_t noise_power,fp_t **q2,int np,fp_t ff)
{
  fp_t hi=1.0,lo=0.2;
  fp_t **pt=ft2dim(1,np,1,np);
  for(int x=1;x<=np;++x)
    for(int y=1;y<=np;++y)
      pt[x][y]=(q2[x][y])?(obj_f[x][y].re*obj_f[x][y].re+obj_f[x][y].im*obj_f[x][y].im)/q2[x][y]:0.0;
  smooth(pt,np,np,3);
  for(int x=1;x<=np;++x)
    for(int y=1;y<=np;++y)
      pt[x][y]=np*np*noise_power/pt[x][y]; 
  fp_t **rr=ft2dim(1,np,1,np);
  for(int x=1;x<=np;++x){
    rr[x][1]=pt[x][1];
    for(int y=2;y<=np;++y) rr[x][y]=pt[x][np-y+2];
  }
  fp_t *tmp=new fp_t [np]-1;
  for(int y=1;y<=np;++y){                        // rr==filter
    rr[1][y]=max((fp_t)(1.0-0.5*(rr[1][y]+pt[1][y])),(fp_t)0.0);
    for(int x=2;x<=np;++x) tmp[x]=rr[x][y];       // temporary storage
    for(int x=2;x<=np;++x) rr[x][y]=max((fp_t)(1.0-0.5*(tmp[np-x+2]+pt[x][y])),(fp_t)0.0);
  }
  delete[] (tmp+1);
  for(int x=1;x<=np;++x)
    for(int y=1;y<=np;++y){
      if(rr[x][y]<lo) rr[x][y]=0.0;
      rr[x][y]=min(rr[x][y],1.0);
    }
  rr[np/2+1][np/2+1]=1.0;                   // ; DC gain = 1
  connected_region(rr,np,np,np/2+1,np/2+1); // all continuous nonzero points from np/2,np/2  
  smooth(rr,np,np,9);
  for(int x=1;x<=np;++x)
    for(int y=1;y<=np;++y)
      if((sqr(x-np/2-1)+sqr(y-np/2-1))>sqr(ff)){
        obj_f[x][y].re=obj_f[x][y].im=0.0;
      }else{
        obj_f[x][y].re*=rr[x][y];
        obj_f[x][y].im*=rr[x][y];
      }
  del_ft2dim(rr,1,np,1,np);
  del_ft2dim(pt,1,np,1,np);
}

void mg_solve(float64_t **xxx,float64_t **rhs,float64_t ****A,int nx,int ny,int level,int offs,float64_t fct)
{
  if(level){
    int sf=2;
    int snx=nx/sf,sny=nx/sf;
    float64_t ****As=f64t4dim(1,snx,1,sny,1,snx,1,sny);
    float64_t **dfct=f64t2dim(1,snx,1,sny);
    float64_t md=0.0,cdm=HUGE_VAL;
    for(int x=1;x<=snx;++x)
      for(int y=1;y<=sny;++y){
        dfct[x][y]=0.0;
        for(int ix=-sf+1;ix<=0;++ix) for(int iy=-sf+1;iy<=0;++iy) dfct[x][y]+=rhs[sf*x+ix][sf*y+iy];
//        dfct[x][y]/=(float64_t)(sf*sf);
        md=max(md,fabs(dfct[x][y]));
        for(int u=1;u<=snx;++u)
          for(int v=1;v<=sny;++v){
            As[x][y][u][v]=0.0;
            for(int ix=-sf+1;ix<=0;++ix) for(int iy=-sf+1;iy<=0;++iy)
              for(int iu=-sf+1;iu<=0;++iu) for(int iv=-sf+1;iv<=0;++iv) As[x][y][u][v]+=A[sf*x+ix][sf*y+iy][sf*u+iu][sf*v+iv];
//            As[x][y][u][v]/=(float64_t)(sf*sf*sf*sf);
          }
        cdm=min(cdm,As[x][y][x][y]);
      }
    fprintf(stderr,"%d: dfct(c) %E, cdm(c)= %E\n",nx,md,cdm);
    float64_t **vvv=f64t2dim(1,snx,1,sny);
    memset(vvv[1]+1,0,snx*sny*sizeof(float64_t));
    for(int l=0;l<=5;++l) mg_solve(vvv,dfct,As,snx,sny,level-1,offs/sf,fct*(float64_t)sf);
    float64_t mv=0.0;
    for(int u=1;u<=snx;++u) for(int v=1;v<=sny;++v){
      mv=max(mv,vvv[u][v]);
      for(int iu=-sf+1;iu<=0;++iu) for(int iv=-sf+1;iv<=0;++iv) xxx[sf*u+iu][sf*v+iv]+=vvv[u][v]/(float64_t)(sf*sf);
    }
    fprintf(stderr,"%d: corr(c) %E\n",nx,mv);
    
    del_f64t2dim(vvv,1,snx,1,sny);
    del_f64t2dim(dfct,1,snx,1,sny);
    del_f64t4dim(As,1,snx,1,sny,1,snx,1,sny);
  }
  float64_t **dfct=f64t2dim(1,nx,1,ny);
  float64_t md=0.0;
  for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // defect = rhs - A * x;
    dfct[x][y]=rhs[x][y];
    for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) dfct[x][y]-=A[x][y][u][v]*xxx[u][v];
    md=max(md,fabs(dfct[x][y]));
  }
  fprintf(stderr,"%d: dfct(f) %E\n",nx,md);
  for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) xxx[u][v]+=fct*dfct[u][v]/A[u][v][u][v];
  del_f64t2dim(dfct,1,nx,1,ny);
// wipe over the boundaries (to avoid ringing)
  for(int u=1;u<=offs;++u)       for(int v=1+offs;v<=ny-offs;++v) xxx[u][v]=xxx[1+offs][v];
  for(int u=nx-offs+1;u<=nx;++u) for(int v=1+offs;v<=ny-offs;++v) xxx[u][v]=xxx[nx-offs][v];
  for(int u=1;u<=nx;++u)         for(int v=1;v<=offs;++v)         xxx[u][v]=xxx[u][1+offs];
  for(int u=1;u<=nx;++u)         for(int v=ny-offs+1;v<=ny;++v)   xxx[u][v]=xxx[u][ny-offs];
}

void orthomin(int iter,float64_t **p,int nx,int ny,float64_t ***ap,float64_t &ap2nor)
{
  int itr=iter-5;
  itr=(itr%10);
//  fprintf(stderr,"OM: %d %d\n",iter,itr);
  if(itr>5) itr=0;
  if(itr==1){
    for(int x=1;x<=nx;++x)
      for(int y=1;y<=ny;++y)
        ap[x][y][1]=p[x][y];
    return;
  }else if(itr==2){
    for(int x=1;x<=nx;++x)
      for(int y=1;y<=ny;++y){
        ap[x][y][6]=1.0/p[x][y];
        ap[x][y][3]=p[x][y]-ap[x][y][1];
        ap[x][y][4]=ap[x][y][3];
        ap[x][y][1]=p[x][y];
        ap[x][y][0]=p[x][y];
      }
    return;
  }else if(itr<=0) return;
  for(int x=1;x<=nx;++x)                           // compute conjugate vector P and AP
    for(int y=1;y<=ny;++y)
      ap[x][y][1]=ap[x][y][4]-(p[x][y]-ap[x][y][1]);
  if(itr>3){                                      // new coeffcient b - eq. 75
    float64_t bb=0.0;
    for(int x=1;x<=nx;++x)
      for(int y=1;y<=ny;++y)
        bb+=ap[x][y][1]*ap[x][y][2]*ap[x][y][6];
    bb=bb/ap2nor;
    for(int x=1;x<=nx;++x)
      for(int y=1;y<=ny;++y){                      // new P and AP - eqs. 74 and 76
        ap[x][y][4]-=bb*ap[x][y][5];
        ap[x][y][1]-=bb*ap[x][y][2];
      }
  }
  float64_t ap1nor=0.0;
  float64_t aa=0.0;
  for(int x=1;x<=nx;++x)
    for(int y=1;y<=ny;++y){
      ap1nor+=ap[x][y][1]*ap[x][y][1]*ap[x][y][6];
      aa+=ap[x][y][3]*ap[x][y][1]*ap[x][y][6];
    }
  aa/=ap1nor;
//  fprintf(stderr,"OM %d!\n",itr);
  for(int x=1;x<=nx;++x)
    for(int y=1;y<=ny;++y){
      p[x][y]=ap[x][y][0]+aa*(ap[x][y][4]-ap[x][y][1]);  // accelerated iterate - eq. 78
      ap[x][y][3]-=aa*ap[x][y][1];
      ap[x][y][2]=ap[x][y][1];                           // store quantities for use in the following iterations
      ap[x][y][0]=p[x][y];
      ap[x][y][1]=p[x][y];
      ap[x][y][5]=ap[x][y][4];
      ap[x][y][4]=ap[x][y][3];
  }
  ap2nor=ap1nor;
}

#define THRD_CMD_RD   0x01
#define THRD_CMD_RMP  0x02
#define THRD_CMD_MM   0x03
#define THRD_CMD_DOT  0x04
#define THRD_CMD_SLV  0x05
#define THRD_CMD_SLV2 0x06
#define THRD_CMD_SLV3 0x07
#define THRD_CMD_QUIT 0xFF

#define SLV_METHOD_DEF   0
#define SLV_METHOD_MG    1
#define SLV_METHOD_LR    2

struct mm{
  momfbd_struct &mimg;
  int nframes;
  
  float64_t *****M,****MM,**bb,***b;
  mm(momfbd_struct &mimg_in,int nframes_in):mimg(mimg_in),nframes(nframes_in)
  {
  };
  
};

//void *matmul(struct mm *m,int x1,int y1);

struct ptsk{
  struct mm *m;
  int i,x,y,npts,nx,ny;
  float64_t tx,ty;
  int dxl,dxh,dyl,dyh;
  int ll,lh;
//
  float64_t ****rmap;
  float64_t ***a,**b;
  float64_t ****A,**Ap;
  float64_t ***bb,***xx;
  float64_t ***rawscan;
  float64_t cdm;
  uint08_t **mask;
  int mul;
  int *idx,*pos;
  int **xl,**xh,**yl,**yh;
  int niter;
//
  char *rawpath;
//
  io_class *io;
//  
  ptsk(void){};
  ptsk(struct mm *m_in,int i_in,int npts_in,io_class *io_in){
    m=m_in;
    i=i_in;
    npts=npts_in;
    io=io_in;
  };
  ptsk(struct mm *m_in,int i_in,int npts_in,float64_t ****rmap_in,float64_t ***rawscan_in,io_class *io_in){
    m=m_in;
    i=i_in;
    npts=npts_in;
    rmap=rmap_in;
    rawscan=rawscan_in;
    io=io_in;
  };
  ptsk(struct mm *m_in,int x_in,int y_in,int npts_in,float64_t ****rmap_in,float64_t ****A_in,io_class *io_in){
    m=m_in;
    x=x_in;
    y=y_in;
    npts=npts_in;
    rmap=rmap_in;
    A=A_in;
    io=io_in;
  };
  ptsk(struct mm *m_in,int x_in,int y_in,int npts_in,float64_t ****A_in,float64_t ***a_in,float64_t **b_in,io_class *io_in){
    m=m_in;
    x=x_in;
    y=y_in;
    npts=npts_in;
    A=A_in;
    a=a_in;
    b=b_in;
    io=io_in;
  };
  ptsk(struct mm *m_in,int x_in,int y_in,int npts_in,float64_t ****A_in,float64_t ***xx_in,float64_t ***bb_in,io_class *io_in){
    m=m_in;
    x=x_in;
    y=y_in;
    npts=npts_in;
    A=A_in;
    xx=xx_in;
    bb=bb_in;
    io=io_in;
  };
  ptsk(struct mm *m_in,int x_in,int y_in,int npts_in,float64_t ****rmap_in,float64_t ***xx_in,float64_t ***bb_in,float64_t **ff,io_class *io_in){
    m=m_in;
    npts=npts_in;
    rmap=rmap_in;
    xx=xx_in;
    bb=bb_in;
    b=ff;
    io=io_in;
  };
  ptsk(struct mm *m_in,float64_t ****A_in,float64_t **Ap_in,int *idx_in,int *pos_in,io_class *io_in){
    m=m_in;
    A=A_in;
    Ap=Ap_in;
    idx=idx_in;
    pos=pos_in;
    io=io_in;
  };
};

class thrd_class{
  int spair[2];
  pthread_t thrd;
  int idle;
public:
  thrd_class(fdset&);
  ~thrd_class(void);
  void *run(void);
  int send(uint08_t,class ptsk&);
  int recv(uint08_t);
  int recv(float64_t**&);
  int recv(float64_t***&);
};

void *thrd_func(void *arg){
  return ((thrd_class*)arg)->run();
}

thrd_class::thrd_class(fdset &fds)
{
  if(socketpair(AF_UNIX,SOCK_STREAM,0,spair)<0){
    fprintf(stderr,"thrd_class::thrd_class: failed to create socket pair: %s\n",strerror(errno));
    exit(1);
  }
  fds.add(spair[0],this);
  if(pthread_create(&(thrd),0,thrd_func,this)<0){
    fprintf(stderr,"thrd_class::thrd_class: failed to create thread: %s\n",strerror(errno));
    exit(1);
  }
  idle=1;
}

thrd_class::~thrd_class(void)
{
  void *rv;
  uint08_t cmd=THRD_CMD_QUIT;
  if(::send(spair[0],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::~thrd_class: failed to send stop command: %s\n",strerror(errno));
  if(pthread_join(thrd,&rv)<0) fprintf(stderr,"thrd_class::~thrd_class: failed to join thread: %s\n",strerror(errno));
  close(spair[0]);
}

int thrd_class::send(uint08_t cmd,class ptsk &tsk)
{
  if(idle){
     idle=0;
     if(::send(spair[0],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::send: failed to send command: %s\n",strerror(errno));
     if(::send(spair[0],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::~thrd_send: failed to send task: %s\n",strerror(errno));
     return 1;
  }
  return 0;
}

int thrd_class::recv(uint08_t cmd)
{
  if(!idle){
    uint08_t cmd_o;
    if(::recv(spair[0],&cmd_o,1,0)<=0) fprintf(stderr,"thrd_class::recv: failed to receive command: %s\n",strerror(errno));
    if(cmd!=cmd_o) fprintf(stderr,"thrd_class::recv: received command 0x%02X differs from expected command 0x%02X\n",cmd_o,cmd);
    idle=1;
    return 1;
  }
  return 0;
}

int thrd_class::recv(float64_t **&p)
{
  if(::recv(spair[0],&p,sizeof(float64_t**),0)<=0) fprintf(stderr,"thrd_class::recv: failed to receive address: %s\n",strerror(errno));
  return 0;
}

int thrd_class::recv(float64_t ***&p)
{
  if(::recv(spair[0],&p,sizeof(float64_t***),0)<=0) fprintf(stderr,"thrd_class::recv: failed to receive address: %s\n",strerror(errno));
  return 0;
}

float64_t mul(float64_t **r1,float64_t **r2,int xl,int xh,int yl,int yh)
{ // in this bit, the coordinates were reversed
  float64_t rv=0.0;
  for(int x=xl;x<=xh;++x) for(int y=yl;y<=yh;++y) rv+=r1[x][y]*r2[x][y];
//  for(int x=xl;x<=xh;++x) for(int y=yl;y<=yh;++y) rv+=r1[x][y]*r1[x][y];
  return rv;
}

void *thrd_class::run(void)
{
  int done=0;
  while(!done){
    uint08_t cmd;
    if(::recv(spair[1],&cmd,1,0)>0){
      switch(cmd){
        case(THRD_CMD_QUIT):{
          close(spair[1]);
          return 0;
        }
        case(THRD_CMD_RD):{
          struct ptsk tsk;
          if(::recv(spair[1],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::run: failed to receive task: %s\n",strerror(errno));
// 
          int i=tsk.i;
          int ll=tsk.ll,lh=tsk.lh;
          int mul=tsk.mul;

          int fidx=i-1; // index of this raw frame
          char *raw_name=(tsk.m->mimg.get_names())[fidx];
//
          int nx=tsk.nx,ny=tsk.ny;
          int dxl=tsk.dxl,dxh=tsk.dxh,dyl=tsk.dyl,dyh=tsk.dyh; // data coordinates...
//   data coordinates
//          if(i==-1) tsk.io->msg(IOL_INFO,"patch: %d %d %d %d %d %d\n",xs,ys,tsk.xc,tsk.yc,nx,ny);
//
          int sx,sy;
          tsk.m->mimg.centerpatch(nx/2,ny/2,sx,sy);
//
          float64_t ***rawscan=f64t3dim(dxl,dxh,dyl,dyh,ll,lh);
          memset(rawscan[dxl][dyl]+ll,0,(dxh-dxl+1)*(dyh-dyl+1)*(lh-ll+1)*sizeof(float64_t));
//
          float64_t ***mask=0;
          {
            int *ds=0,ndim,type;
            char *header;
            if(byte *tmp=ana_fzread((char*)"mask.fillpix.f0",ds,ndim,header,type)){// file was found
              mask=f64t3dim((float64_t*)tmp,1,ds[3],dyl,dyh,dxl,dxh);
              if(header) delete[] header;
              delete[] (ds+1);
            }
          }
// read the data
          if(raw_name){
            char *tmpname=strcpy(new char [strlen(raw_name)+1],raw_name);
            int *ds=0,ndim,type;
            char *header;
            byte *tmp=0;
            if(0) tmp=ana_fzread(tmpname,ds,ndim,header,type);
            if(!tmp){ // file not found
              char *filename=strrchr(tmpname,'/');
              filename[0]=0;
              filename+=1;
              char *basepath=strrchr(tmpname,'/');
              basepath[0]=0;
              char path[1000];
              sprintf(path,"%s/%s",tsk.rawpath,filename);
//              tmp=ana_fzread(path,ds,ndim,header,type);
              if(!(tmp=ana_fzread(path,ds,ndim,header,type))){// file not found
                char rpath[1000];
                sprintf(rpath,"%s/%s/%s",tmpname,tsk.rawpath,filename);
                tsk.io->msg(IOL_WARN,"attempting to fetch input file %s\n",filename);
                char *scmd=new char [1000];
                sprintf(scmd,"scp vannoort@gofio2.mps.mpg.de:%s %s/\n",rpath,tsk.rawpath);
                if(system(scmd)){
                  sleep(1);
                  if(system(scmd)) tsk.io->msg(IOL_ERROR,"failed to copy input file %s\n",raw_name);
                }
                if(!(tmp=ana_fzread(path,ds,ndim,header,type))){// file really not found
                  tsk.io->msg(IOL_ERROR,"failed to find input file %s: this will cause problems!\n",path);
                }
                delete[] scmd;
              }
              if(ndim!=3) tsk.io->msg(IOL_ERROR,"File %s had %d dimensions, expected 3!\n",path,ndim);
            }else{
              if(ndim!=3) tsk.io->msg(IOL_ERROR,"File %s had %d dimensions, expected 3!\n",tmpname,ndim);
            }
            if(tmp){ // file read succesfully
              tsk.io->msg(IOL_DEB1,"File %s (%016lX) has %d dimensions: %d %d %d (%d %d %d)\n",raw_name,(tmp),ndim,ds[1],ds[2],ds[3],dxh-dxl+1,dyh-dyl+1,ds[3]);
              switch(type){
                case(FLOAT64):{
                  float64_t ***im=f64t3dim((float64_t*)tmp,1,ds[3],dyl,dyh,dxl,dxh);
                  for(int l=ll;l<=lh;++l){
                    if(mask) for(int x=dxl;x<=dxh;++x) for(int y=dyl;y<=dyh;++y) if(!mask[l][y][x]) im[l][y][x]=0.0;
                    im[l]=fillpix(im[l],dyl,dyh,dxl,dxh,CFG_FPM_MEDIAN,dyl,dyh,dxl,dxh);
                    for(int x=dxl;x<=dxh;++x)
                      for(int y=dyl;y<=dyh;++y) rawscan[x][y][l]+=(float64_t)(im[l][y][x]);
                  }
                  del_f64t3dim(im,1,ds[3],dyl,dyh,dxl,dxh);
                  break;
                }
                case(FLOAT32):{
                  float32_t ***im=f32t3dim((float32_t*)tmp,1,ds[3],dyl,dyh,dxl,dxh);
                  for(int l=ll;l<=lh;++l){
                    if(mask) for(int x=dxl;x<=dxh;++x) for(int y=dyl;y<=dyh;++y) if(!mask[l][y][x]) im[l][y][x]=0.0;
                    im[l]=fillpix(im[l],dyl,dyh,dxl,dxh,CFG_FPM_MEDIAN,dyl,dyh,dxl,dxh);
                    for(int x=dxl;x<=dxh;++x)
                      for(int y=dyl;y<=dyh;++y) rawscan[x][y][l]+=(float64_t)(im[l][y][x]);
                  }
                  del_f32t3dim(im,1,ds[3],dyl,dyh,dxl,dxh);
                  break;
                }
                default: tsk.io->msg(IOL_ERROR,"File %s had unknown type %d!\n",tmpname,type);
              }
            }
            if(header) delete[] header;
            delete[] (ds+1);
          }
//
          if(i==-1){
              tsk.io->msg(IOL_INFO,"writing raw scan!\n");
              int ds[]={0,lh-ll+1,dyh-dyl+1,dxh-dxl+1};
              char name[100];
              sprintf(name,"rawscan.f0");
              ana_fzwrite((byte*)(rawscan[dxl][dyl]+ll),name,ds,3,(char*)"bla",FLOAT64);
          }
          
          float64_t ***rhs_i=f64t3dim(1,nx,1,ny,ll,lh);
          memset(rhs_i[1][1]+ll,0,nx*ny*(lh-ll+1)*sizeof(float64_t));
          float64_t **ff_i=f64t2dim(1,nx,1,ny);
          memset(ff_i[1]+1,0,nx*ny*sizeof(float64_t));
//
          for(int iy=1;iy<=ny;++iy) // loop over solar y coordinate
            for(int ix=1;ix<=nx;++ix){ // loop over solar x coordinate
              int np;
//
              float64_t xp=tsk.tx,yp=tsk.ty; // coordinates for object at x/ypos[nd/2]
              float64_t **psf=tsk.m->mimg.psf(ix,iy,i,mul,xp,yp,np,*tsk.io);
//
              int dx=ix-np/2,dy=iy-np/2;
              int xl=max(1+dx-(int)round(xp),dxl),xh=min(np+dx-(int)round(xp),dxh);
              int yl=max(1+dy-(int)round(yp),dyl),yh=min(np+dy-(int)round(yp),dyh);
              if(tsk.mask){
                for(int x=xl;x<=xh;++x) // for each solar coordinate, add contribution to data location and multiply with transpose(A)
                  for(int y=yl;y<=yh;++y)
                    if(tsk.mask[x][y]){
                      float64_t f=psf[x-dx+(int)round(xp)][y-dy+(int)round(yp)];
                      for(int l=ll;l<=lh;++l) rhs_i[ix][iy][l]+=f*rawscan[x][y][l];
                      ff_i[ix][iy]+=f;
                    }
              }else{
                for(int x=xl;x<=xh;++x) // for each solar coordinate, add contribution to data location and multiply with transpose(A)
                  for(int y=yl;y<=yh;++y){
                    float64_t f=psf[x-dx+(int)round(xp)][y-dy+(int)round(yp)];
                    for(int l=ll;l<=lh;++l) rhs_i[ix][iy][l]+=f*rawscan[x][y][l];
                    ff_i[ix][iy]+=f;
                  }
              }
              del_f64t2dim(psf,1,np,1,np);
            }
          del_f64t3dim(rawscan,dxl,dxh,dyl,dyh,ll,lh);
          if(i==-1){
             tsk.io->msg(IOL_INFO,"writing rhs_i and rhsff_i!\n");
             int ds[]={0,lh-ll+1,ny,nx};
              char name[100];
              sprintf(name,"rhs_i.f0");
              ana_fzwrite((byte*)(rhs_i[1][1]+ll),name,ds,3,(char*)"bla",FLOAT64);
              int dds[]={0,ny,nx};
              sprintf(name,"rhsff_i.f0");
              ana_fzwrite((byte*)(ff_i[1]+1),name,dds,2,(char*)"bla",FLOAT64);
          }
//
          if(::send(spair[1],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::run: failed to return command: %s\n",strerror(errno));
          if(::send(spair[1],&rhs_i,sizeof(float64_t***),0)<=0) fprintf(stderr,"thrd_class::run: failed to return rhs: %s\n",strerror(errno));
          if(::send(spair[1],&ff_i,sizeof(float64_t**),0)<=0) fprintf(stderr,"thrd_class::run: failed to return ff: %s\n",strerror(errno));
          break;
        }
        case(THRD_CMD_RMP):{
          struct ptsk tsk;
          if(::recv(spair[1],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::run: failed to receive task: %s\n",strerror(errno));
// data frame
          int i=tsk.i;
          int mul=tsk.mul;
// solar coordinates 
          int iy=tsk.y,ix=tsk.x;
// dims
          int npts=tsk.npts;
          int nx=tsk.nx,ny=tsk.ny;
          int dxl=tsk.dxl,dxh=tsk.dxh,dyl=tsk.dyl,dyh=tsk.dyh;
// local stuff
          int np;
          float64_t xp=tsk.tx,yp=tsk.ty; // coordinates for object at x,y
          float64_t **psf=tsk.m->mimg.psf(ix,iy,i,mul,xp,yp,np,*tsk.io);
//
          int dx=ix-np/2,dy=iy-np/2;
          int xl=max(1+dx-(int)round(xp),dxl),xh=min(np+dx-(int)round(xp),dxh);
          int yl=max(1+dy-(int)round(yp),dyl),yh=min(np+dy-(int)round(yp),dyh);
// save PSF in "reversed" location
          if(tsk.mask){
            for(int x=xl;x<=xh;++x) // for each solar coordinate, add contribution to data location (x,y)
              for(int y=yl;y<=yh;++y)
                if(tsk.mask[x][y]) tsk.rmap[ix][iy][x][y]+=psf[x-dx+(int)round(xp)][y-dy+(int)round(yp)];
          }else{
            for(int x=xl;x<=xh;++x) // for each solar coordinate, add contribution to data location (x,y) 
              for(int y=yl;y<=yh;++y)
                tsk.rmap[ix][iy][x][y]+=psf[x-dx+(int)round(xp)][y-dy+(int)round(yp)];
          }
          del_f64t2dim(psf,1,np,1,np);
//
          tsk.xl[ix][iy]=xl;
          tsk.xh[ix][iy]=xh;
          tsk.yl[ix][iy]=yl;
          tsk.yh[ix][iy]=yh;
//
          if(::send(spair[1],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::run: failed to return command: %s\n",strerror(errno));
          break;
        }
        case(THRD_CMD_MM):{
          struct ptsk tsk;
          if(::recv(spair[1],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::run: failed to receive task: %s\n",strerror(errno));
          int x=tsk.x,y=tsk.y,np=tsk.npts;
          int nx=tsk.nx,ny=tsk.ny;
          float64_t **a=tsk.A[x][y];  // row
// make use of the symmetry of A: only calculate the upper triangle
          for(int v=y;v<=ny;++v){     // column
            int yl=max(tsk.yl[x][y],tsk.yl[x][v]),yh=min(tsk.yh[x][y],tsk.yh[x][v]);
            int xl=max(tsk.xl[x][y],tsk.xl[x][v]),xh=min(tsk.xh[x][y],tsk.xh[x][v]);
            a[x][v]+=mul(tsk.rmap[x][y],tsk.rmap[x][v],xl,xh,yl,yh);
          }
          for(int u=x+1;u<=nx;++u)    // column
            for(int v=1;v<=ny;++v){
              int yl=max(tsk.yl[x][y],tsk.yl[u][v]),yh=min(tsk.yh[x][y],tsk.yh[u][v]);
              int xl=max(tsk.xl[x][y],tsk.xl[u][v]),xh=min(tsk.xh[x][y],tsk.xh[u][v]);
              a[u][v]+=mul(tsk.rmap[x][y],tsk.rmap[u][v],xl,xh,yl,yh);
            }
//
          if(::send(spair[1],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::run: failed to return command: %s\n",strerror(errno));
          break;
        }
        case(THRD_CMD_DOT):{ // evaluate the reduced forward problem
          struct ptsk tsk;
          if(::recv(spair[1],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::run: failed to receive task: %s\n",strerror(errno));
// 
          int x=tsk.x,y=tsk.y;
          int nx=tsk.nx,ny=tsk.ny;
          tsk.b[x][y]=0.0;
          for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) tsk.b[x][y]+=tsk.A[x][y][u][v]*tsk.a[0][u][v];
//
          if(::send(spair[1],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::run: failed to return command: %s\n",strerror(errno));
          break;
        }
        case(THRD_CMD_SLV):{ // evaluate A*xx == bb
          struct ptsk tsk;
          if(::recv(spair[1],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::run: failed to receive task: %s\n",strerror(errno));
// 
          int nx=tsk.nx,ny=tsk.ny,l=tsk.i;
          
          float64_t ****A=tsk.A,**xx=tsk.xx[l],**rhs=tsk.bb[l],cdm=tsk.cdm;

          float64_t **dfct=f64t2dim(1,nx,1,ny);
          float64_t **vvv=f64t2dim(1,nx,1,ny);
//          for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) xx[x][y]=600.0; // start flat
//
          int offs=8; // embed restored image to minimize edge effects?
//          char name[1000];
//          sprintf(name,"bla/md.%d.txt",i);
//          FILE *f=fopen(name,"w");
//            mgsolve(A,rhs,xx,npts);
          int niter=max(12,(int)round(8.0/sqrt(cdm)));
//          int niter=max(16,(int)round(16.0/sqrt(cdm)));
          float64_t fct=min(0.02,0.025*sqrt(cdm));
          for(int c=0;c<=niter;++c){
            for(int k=0;k<=8;++k){
              for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // defect = rhs - A * x;
                dfct[x][y]=rhs[x][y];
                for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) dfct[x][y]-=A[x][y][u][v]*xx[u][v];
              }
//
              int seq[]={16,8,4,2,1,2,4,8,16};
              int sf=seq[k];
//
              float64_t md=0.0;
              for(int u=1+offs;u<=nx-offs;++u) for(int v=1+offs;v<=ny-offs;++v) md=max(fabs(dfct[u][v]),md);
              fprintf(stderr,"%d %d %d %E\n",c,k,sf,md);
//
              if(sf>1){
                for(int u=1;u<=nx/sf;++u)
                  for(int v=1;v<=ny/sf;++v){ // solve the coarse problem
                    float64_t a=0.0,d=0.0;
                    for(int u1=-sf+1;u1<=0;++u1)
                      for(int v1=-sf+1;v1<=0;++v1){
                        d+=dfct[sf*u+u1][sf*v+v1];
                        for(int u2=-sf+1;u2<=0;++u2)
                          for(int v2=-sf+1;v2<=0;++v2) a+=A[sf*u+u1][sf*v+v1][sf*u+u2][sf*v+v2];
                      }
                    for(int u1=-sf+1;u1<=0;++u1) for(int v1=-sf+1;v1<=0;++v1) vvv[sf*u+u1][sf*v+v1]=(float64_t)sf*d/a;
                  }
              }else{
                for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) vvv[u][v]=dfct[u][v]/A[u][v][u][v];
              }
              for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) xx[u][v]+=fct*vvv[u][v];
// wipe over the boundaries (to avoid ringing)
              for(int u=1;u<=offs;++u)       for(int v=1+offs;v<=ny-offs;++v) xx[u][v]=xx[1+offs][v];
              for(int u=nx-offs+1;u<=nx;++u) for(int v=1+offs;v<=ny-offs;++v) xx[u][v]=xx[nx-offs][v];
              for(int u=1;u<=nx;++u)         for(int v=1;v<=offs;++v)         xx[u][v]=xx[u][1+offs];
              for(int u=1;u<=nx;++u)         for(int v=ny-offs+1;v<=ny;++v)   xx[u][v]=xx[u][ny-offs];
            }
          }
/*
          fct=0.015;
          for(int c=0;c<=100;++c){
            for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // defect = rhs - A * x;
              dfct[x][y]=rhs[x][y];
              for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) dfct[x][y]-=A[x][y][u][v]*xx[u][v];
            }
            float64_t md=0.0;
            for(int u=1+offs;u<=nx-offs;++u) for(int v=1+offs;v<=ny-offs;++v) md=max(fabs(dfct[u][v]),md);
            fprintf(stderr,"%d %E\n",c,md);
            for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) vvv[u][v]=dfct[u][v]/A[u][v][u][v];
            for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) xx[u][v]+=fct*vvv[u][v];
// wipe over the boundaries (to avoid ringing)
            for(int u=1;u<=offs;++u)       for(int v=1+offs;v<=ny-offs;++v) xx[u][v]=xx[1+offs][v];
            for(int u=nx-offs+1;u<=nx;++u) for(int v=1+offs;v<=ny-offs;++v) xx[u][v]=xx[nx-offs][v];
            for(int u=1;u<=nx;++u)         for(int v=1;v<=offs;++v)         xx[u][v]=xx[u][1+offs];
            for(int u=1;u<=nx;++u)         for(int v=ny-offs+1;v<=ny;++v)   xx[u][v]=xx[u][ny-offs];
          }*/
//          fclose(f);
          del_f64t2dim(vvv,1,nx,1,ny);
          del_f64t2dim(dfct,1,nx,1,ny);
          if(::send(spair[1],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::run: failed to return command: %s\n",strerror(errno));
          break;
        }
        case(THRD_CMD_SLV2):{ // evaluate A*xx == bb
          struct ptsk tsk;
          if(::recv(spair[1],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::run: failed to receive task: %s\n",strerror(errno));
// 
          int l=tsk.i,nx=tsk.nx,ny=tsk.ny;
          
          float64_t ****A=tsk.A,**xxx=tsk.xx[l],**rhs=tsk.bb[l],cdm=tsk.cdm;

          float64_t **dfct=f64t2dim(1,nx,1,ny);
          float64_t **vvv=f64t2dim(1,nx,1,ny);
//
          int offs=8; // embed restored image to minimize edge effects?
          int niter=max(16,(int)round(32.0/sqrt(cdm)));
//
          for(int c=0;c<=niter;++c){
            float64_t md=0.0;
            for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // defect = rhs - A * x;
              dfct[x][y]=rhs[x][y];
              for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) dfct[x][y]-=A[x][y][u][v]*xxx[u][v];
              md=max(md,fabs(dfct[x][y]));
            }
            fprintf(stderr,"dfct: %d %E\n",c,md);
            memset(vvv[1]+1,0,nx*ny*sizeof(float64_t));
            mg_solve(vvv,dfct,A,nx,ny,2,offs,1E-3);
            float64_t mv=0;
            for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v){
              mv=max(mv,fabs(vvv[u][v]));
              xxx[u][v]+=vvv[u][v];
            }
            fprintf(stderr,"corr: %d %E\n",c,mv);
//
            for(int w=0;w<=3;++w){
              for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // defect = rhs - A * x;
                dfct[x][y]=rhs[x][y];
                for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) dfct[x][y]-=A[x][y][u][v]*xxx[u][v];
              }
              for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) xxx[x][y]+=1E-3*dfct[x][y]/A[x][y][x][y];
            }
//
          }
          del_f64t2dim(vvv,1,nx,1,ny);
          del_f64t2dim(dfct,1,nx,1,ny);
          if(::send(spair[1],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::run: failed to return command: %s\n",strerror(errno));
          break;
        }
        case(THRD_CMD_SLV3):{ // variation on Lucy-Richardson: this will only work for LC states!
          struct ptsk tsk;
          if(::recv(spair[1],&tsk,sizeof(struct ptsk),0)<=0) fprintf(stderr,"thrd_class::run: failed to receive task: %s\n",strerror(errno));
// 
          int nx=tsk.nx,ny=tsk.ny,l=tsk.i;
          
          float64_t ****A=tsk.A,**xxx=tsk.xx[l],cdm=tsk.cdm;

          float64_t **dfct=f64t2dim(1,nx,1,ny);
          float64_t **vvv=f64t2dim(1,nx,1,ny);
          float64_t **rhs=tsk.bb[l];
//          float64_t **rhs=f64t2dim(1,nx,1,ny);
//          for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) rhs[x][y]=tsk.b[x][y];
//          
          int16_t **ul=i16t2dim(1,nx,1,ny),**uh=i16t2dim(1,nx,1,ny);
          int16_t **vl=i16t2dim(1,nx,1,ny),**vh=i16t2dim(1,nx,1,ny);
// avoid unnecessary work!
          for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){
            ul[x][y]=nx;
            vl[x][y]=ny;
            uh[x][y]=1;
            vh[x][y]=1;
            for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) if(A[x][y][u][v]){
              ul[x][y]=min(ul[x][y],u);
              uh[x][y]=max(uh[x][y],u);
              vl[x][y]=min(vl[x][y],v);
              vh[x][y]=max(vh[x][y],v);
            }
          }
//
          float64_t ap2nor,***ap=f64t3dim(1,nx,1,ny,0,6);
//
          int offs=8; // embed restored image to minimize edge effects?
          for(int c=0;c<=tsk.niter;++c){
            for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // A * x
              dfct[x][y]=0.0;
              for(int16_t u=ul[x][y];u<=uh[x][y];++u) for(int16_t v=vl[x][y];v<=vh[x][y];++v) dfct[x][y]+=A[x][y][u][v]*xxx[u][v];
              dfct[x][y]=rhs[x][y]/dfct[x][y];
            }
            for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // (rhs/ A * x) * A
              vvv[x][y]=0.0;
              for(int16_t u=ul[x][y];u<=uh[x][y];++u) for(int16_t v=vl[x][y];v<=vh[x][y];++v) vvv[x][y]+=A[x][y][u][v]*dfct[u][v];
            }
            for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) xxx[x][y]*=vvv[x][y];
// wipe over the boundaries (to avoid ringing)
            for(int u=1;u<=offs;++u)       for(int v=1+offs;v<=ny-offs;++v) xxx[u][v]=xxx[1+offs][v];
            for(int u=nx-offs+1;u<=nx;++u) for(int v=1+offs;v<=ny-offs;++v) xxx[u][v]=xxx[nx-offs][v];
            for(int u=1;u<=nx;++u)         for(int v=1;v<=offs;++v)         xxx[u][v]=xxx[u][1+offs];
            for(int u=1;u<=nx;++u)         for(int v=ny-offs+1;v<=ny;++v)   xxx[u][v]=xxx[u][ny-offs];
// apply some ad-oc accelleration: gives about x10 speedup :)
            orthomin(c,xxx,nx,ny,ap,ap2nor);
          }
//
          del_f64t3dim(ap,1,nx,1,ny,0,6);
//
          del_i16t2dim(ul,1,nx,1,ny);
          del_i16t2dim(uh,1,nx,1,ny);
          del_i16t2dim(vl,1,nx,1,ny);
          del_i16t2dim(vh,1,nx,1,ny);
//
          del_f64t2dim(vvv,1,nx,1,ny);
          del_f64t2dim(dfct,1,nx,1,ny);
          if(::send(spair[1],&cmd,1,0)<=0) fprintf(stderr,"thrd_class::run: failed to return command: %s\n",strerror(errno));
          break;
        }
        default:{
          fprintf(stderr,"thrd_class::run: unknown command: 0x%02x\n",cmd);
        }
      }
    }else fprintf(stderr,"thrd_class::run: failed to receive command: %s\n",strerror(errno));
  }
  return 0;
}

class mt_class{
  thrd_class **thrd;
  int nthrd;
  fdset fds;
public:
  mt_class(int);
  ~mt_class(void);
//
  void next(uint08_t cmd,class ptsk &tsk);
  void wait(uint08_t cmd);
  void next(uint08_t cmd,class ptsk &tsk,float64_t***,float64_t**);
  void wait(uint08_t cmd,int,int,int,int,float64_t***,float64_t**);
//
  void mkrhs(struct mm&,int,float64_t***,float64_t,float64_t,int,int,int,int,int dxl,int dxh,int dyl,int dyh,uint08_t **mask,int mul,int mod,int off,char *rawpath,io_class&);
  void dot(struct mm &m,int,int,float64_t ****A,float64_t **a,float64_t **b,io_class &io);
  void solve(struct mm &m,int nx,int ny,float64_t ****A,float64_t ***b,float64_t ***x,int,int,float64_t cdm,int,int,io_class &io);
  void mkmat(struct mm &m,int npts,float64_t ****A,int nx,int ny,int dxl,int dxh,int dyl,int dyh,uint08_t **mask,float64_t,float64_t,int mul,int mod,int off,io_class &io);
};

mt_class::mt_class(int nthrd_in)
{
  nthrd=nthrd_in;
  thrd=new thrd_class* [nthrd];
  for(int t=0;t<nthrd;++t) thrd[t]=new thrd_class(fds);
}

mt_class::~mt_class(void)
{
  if(thrd){
    for(int t=0;t<nthrd;++t) delete thrd[t];
    delete[] thrd;
  }
}

void mt_class::next(uint08_t cmd,class ptsk &tsk)
{
  for(int t=0;t<nthrd;++t) if(thrd[t]->send(cmd,tsk)) return;
  fds.act();
  for(void* nxt=fds.next();nxt;nxt=fds.next()){
    if(nxt){
      thrd_class *p=(thrd_class*)nxt;
      if(p->recv(cmd)) if(p->send(cmd,tsk)) return;
    }
  }
}

void mt_class::next(uint08_t cmd,class ptsk &tsk,float64_t ***rhs,float64_t **ff)
{
  for(int t=0;t<nthrd;++t) if(thrd[t]->send(cmd,tsk)) return;
  fds.act();
  for(void* nxt=fds.next();nxt;nxt=fds.next()){
    if(nxt){
      thrd_class *p=(thrd_class*)nxt;
      if(p->recv(cmd)){
        switch(cmd){
          case(THRD_CMD_RD):{
            float64_t ***rhs_i,**ff_i;
            p->recv(rhs_i);
            for(int l=tsk.ll;l<=tsk.lh;++l) for(int x=1;x<=tsk.nx;++x) for(int y=1;y<=tsk.ny;++y) rhs[l][x][y]+=rhs_i[x][y][l];
            del_f64t3dim(rhs_i,1,tsk.nx,1,tsk.ny,tsk.ll,tsk.lh);
            p->recv(ff_i);
            for(int x=1;x<=tsk.nx;++x) for(int y=1;y<=tsk.ny;++y) ff[x][y]+=ff_i[x][y];
            del_f64t2dim(ff_i,1,tsk.nx,1,tsk.ny);
            break;
          }
        }
        if(p->send(cmd,tsk)) return;
      }
    }
  }
}

void mt_class::wait(uint08_t cmd)
{
  for(int t=0;t<nthrd;++t) thrd[t]->recv(cmd);
}

void mt_class::wait(uint08_t cmd,int ll,int lh,int nx,int ny,float64_t ***rhs,float64_t **ff)
{
  for(int t=0;t<nthrd;++t)
    if(thrd[t]->recv(cmd)){
      switch(cmd){
        case(THRD_CMD_RD):{
          float64_t ***rhs_i,**ff_i;
          thrd[t]->recv(rhs_i);
          for(int l=ll;l<=lh;++l) for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) rhs[l][x][y]+=rhs_i[x][y][l];
          del_f64t3dim(rhs_i,1,nx,1,ny,ll,lh);
          thrd[t]->recv(ff_i);
          for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) ff[x][y]+=ff_i[x][y];
          del_f64t2dim(ff_i,1,nx,1,ny);
          break;
        }
      }
    }
}

void mt_class::mkrhs(struct mm &m,int npts,float64_t ***rhs,float64_t tx,float64_t ty,int ll,int lh,int nx,int ny,int dxl,int dxh,int dyl,int dyh,uint08_t **mask,int mul,int mod,int off,char *rawpath,io_class &io)
{
  struct ptsk tsk(&m,0,npts,&io);
  tsk.tx=tx;
  tsk.ty=ty;
  tsk.nx=nx;
  tsk.ny=ny;
  tsk.ll=ll;
  tsk.lh=lh;
//
  tsk.dxl=dxl;
  tsk.dxh=dxh;
  tsk.dyl=dyl;
  tsk.dyh=dyh;
  tsk.mask=mask;
  tsk.mul=mul;
  tsk.rawpath=rawpath;
//  
  float64_t **ff=f64t2dim(1,nx,1,ny);
  memset(ff[1]+1,0,nx*ny*sizeof(float64_t));
  memset(rhs[ll][1]+1,0,nx*ny*(lh-ll+1)*sizeof(float64_t));
  for(int i=1;i<=m.nframes;i+=mul)
    if(((i-1)/mul)%mod==off){
      fprintf(stderr,"%5.1f%%\r",100.0*(float)(i-1)/(float)(m.nframes-1));
      tsk.i=i;
      next(THRD_CMD_RD,tsk,rhs,ff);
    }
  wait(THRD_CMD_RD,ll,lh,nx,ny,rhs,ff);
  for(int l=ll;l<=lh;++l) for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) rhs[l][x][y]/=ff[x][y];
  del_f64t2dim(ff,1,nx,1,ny);
}

//#define __TIMER__

void mt_class::mkmat(struct mm &m,int npts,float64_t ****A,int nx,int ny,int dxl,int dxh,int dyl,int dyh,uint08_t **mask,float64_t tx,float64_t ty,int mul,int mod,int off,io_class &io)
{
  float64_t ****rmap=f64t4dim(1,nx,1,ny,1,nx,1,ny);
  struct ptsk tsk(&m,0,0,npts,rmap,A,&io);
  memset(A[1][1][1]+1,0,nx*ny*nx*ny*sizeof(float64_t));
  tsk.nx=nx;
  tsk.ny=ny;
//
  tsk.dxl=dxl;
  tsk.dxh=dxh;
  tsk.dyl=dyl;
  tsk.dyh=dyh;
  tsk.mask=mask;
  tsk.mul=mul;
//
  tsk.xl=i32t2dim(1,nx,1,ny);
  tsk.xh=i32t2dim(1,nx,1,ny);
  tsk.yl=i32t2dim(1,nx,1,ny);
  tsk.yh=i32t2dim(1,nx,1,ny);

  for(int i=1;i<=m.nframes;i+=mul)
    if(((i-1)/mul)%mod==off){
      tsk.i=i;
//
      tsk.tx=tx;
      tsk.ty=ty;
#ifdef __TIMER__
      struct timeval tv;
      gettimeofday(&tv,0);
      fprintf(stderr,"step 1: %ld\n",(int64_t)tv.tv_sec*1000000LL+(int64_t)tv.tv_usec);
#endif
      memset(rmap[1][1][1]+1,0,nx*ny*nx*ny*sizeof(float64_t));
      for(int x=1;x<=nx;++x){ // loop over solar x coordinate
        tsk.x=x;
        for(int y=1;y<=ny;++y){ // loop over solar y coordinate
          tsk.y=y;
          next(THRD_CMD_RMP,tsk);
        }
      }
      wait(THRD_CMD_RMP);
#ifdef __TIMER__
      gettimeofday(&tv,0);
      fprintf(stderr,"step 2: %ld\n",(int64_t)tv.tv_sec*1000000LL+(int64_t)tv.tv_usec);
#endif
//   add CC to matrix
      for(int x=1;x<=nx;++x){ // loop over solar x coordinate
        tsk.x=x;
        for(int y=1;y<=ny;++y){ // loop over solar y coordinate
          tsk.y=y;
          next(THRD_CMD_MM,tsk);
        }
      }
      wait(THRD_CMD_MM);
#ifdef __TIMER__
      gettimeofday(&tv,0);
      fprintf(stderr,"step 3: %ld\n",(int64_t)tv.tv_sec*1000000LL+(int64_t)tv.tv_usec);
#endif
      fprintf(stderr,"%5.1f%%\r",100.0*(float)((i-1)/(float)(m.nframes-1)));
    }
  del_f64t4dim(rmap,1,nx,1,ny,1,ny,1,nx);
  del_i32t2dim(tsk.xl,1,nx,1,ny);
  del_i32t2dim(tsk.xh,1,nx,1,ny);
  del_i32t2dim(tsk.yl,1,nx,1,ny);
  del_i32t2dim(tsk.yh,1,nx,1,ny);
}

void mt_class::dot(struct mm &m,int nx,int ny,float64_t ****A,float64_t **a,float64_t **b,io_class &io)
{ // b=A*a: evaluate forward problem
  struct ptsk tsk(&m,0,0,0,A,&a,b,&io);
  tsk.nx=nx;
  tsk.ny=ny;
  for(int x=1;x<=nx;++x){
    tsk.x=x;
    for(int y=1;y<=ny;++y){
      fprintf(stderr,"%5.1f%%\r",100.0*(float)((x-1)*ny+(y-1))/(float)(nx*ny-1));
      tsk.y=y;
      next(THRD_CMD_DOT,tsk);
    }
  }
  wait(THRD_CMD_DOT);
}

void mt_class::solve(struct mm &m,int nx,int ny,float64_t ****A,float64_t ***b,float64_t ***x,int ll,int lh,float64_t cdm,int method,int niter,io_class &io)
{ // b=transpose(rmap)*a: generate reduced RHS vector
  int cmd[]={THRD_CMD_SLV,THRD_CMD_SLV2,THRD_CMD_SLV3};
  struct ptsk tsk(&m,0,0,0,A,x,b,&io);
  tsk.cdm=cdm;
  tsk.nx=nx;
  tsk.ny=ny;
  tsk.ll=ll;
  tsk.lh=lh;
  tsk.niter=niter;
//  tsk.i=0;
//  next(cmd[method],tsk);
  for(int l=ll;l<=lh;++l){
    tsk.i=l;
    fprintf(stderr,"%5.1f%%\r",100.0*(float)(l-ll)/(float)(lh-ll));
    next(cmd[method],tsk);
  }
  wait(cmd[method]);
}

int main(int argc,char *argv[])
{ 
  const char *options[]=CMDOPTS,*maps[]=CMDMAPS;    // recognised options
  cmdln cfg(argc,argv,options,maps);
  io_class io(cfg);
//
  int nthrd=1;
  cfg.get("-n",nthrd);
//
  int readrhs=0;
  cfg.get("-r",readrhs);
//
  char *rhsname=0;
  cfg.get("--rhsname",rhsname);
//
  int readmat=0;
  cfg.get("-a",readmat);
//
  char *lhsname=0;
  cfg.get("--lhsname",lhsname);
//
  char *imgbase=0;
  cfg.get("--imgbase",imgbase);
//
  char *rawpath=0;
  cfg.get("--raw",rawpath);
//
  int save=0;
  cfg.get("-s",save);
//
  int noff=0;
  cfg.get("--noff",noff);
//
  int nolhs=0;
  cfg.get("--nolhs",nolhs);
//
  int norhs=0;
  cfg.get("--norhs",norhs);
//
  int noslv=0;
  cfg.get("--noslv",noslv);
//
  int mul=1;
  cfg.get("--mul",mul);
  int mod=1;
  cfg.get("--mod",mod);
  int off=0;
  cfg.get("--off",off);
//
  int *lr=0,nlr=-1;
  char *lrstr=0;
  cfg.get("-lr",lrstr);
  if(lrstr){
    get_numbers(lrstr,lr,nlr);
    if(!nlr){
      io.msg(IOL_ERROR,"LR numbers %s did not expand to any numbers\n",lrstr);
    }else{
      char *p=new char [nlr*8];
      p[0]=0;
      for(int j=1;j<=nlr;++j) sprintf(p+strlen(p)," %d",lr[j]);
      io.msg(IOL_INFO,"LR: using %s iterations\n",p);
      delete[] p;
    }
    delete[] lrstr;
  }
  char *lrbase=0;
  cfg.get("--lrbase",lrbase);
//
// TODO: separate restored coordinates from the data coordinates *and* the momfbd coordinates
//
// map mihi restored coordinates. This can be tricky, the output is transposed!
//
  float64_t xoffs=0.0; // default: no offset
  if(cfg.get("-xo",xoffs)) io.msg(IOL_XNFO,"Offset of spectral data w.r.t. context data in x: %.1f pixels.\n",xoffs);
  float64_t yoffs=0.0; // default: no offset
  if(cfg.get("-yo",yoffs)) io.msg(IOL_XNFO,"Offset of spectral data w.r.t. context data in y: %.1f pixels.\n",yoffs);
  float64_t sf=1.0;    // default: no difference
  if(cfg.get("-sf",sf)) io.msg(IOL_INFO,"Ratio of context data image scale to spectral image scale: %.5f.\n",sf);

/*
  float64_t sf=0.830769;   // x_momfbd=(x_restored-xoffs)/sf
#if 1
// NaD 20180817
  float64_t xoffs=419.0;   // x_restored=0 -> x_momfbd=xoffs
  float64_t yoffs=440.0;   // y_restored=0 -> y_momfbd=yoffs
#else
// Ha 20180824
  float64_t xoffs=419.0;   // x_restored=0 -> x_momfbd=xoffs
  float64_t yoffs=435.0;   // y_restored=0 -> y_momfbd=yoffs
#endif
*/
  momfbd_struct *mimg=0;
  if((!readrhs)||(!readmat)){ // raw data required!
    mimg=new momfbd_struct(argv[1],io);
    mimg->map(-xoffs*sf,-yoffs*sf,sf,io);
//
    mimg->info(io);
  }
#if 1
  if((!readrhs)&&(!norhs)){ // raw data required!
    char **dll=new char* [1];
    dll[0]=0;
    dll=mimg->fetchdata(dll,mul,rawpath,io);
//
    if(dll[0]){
      char *dlpath=mimg->fetchdatapath(io);
      FILE *ff=fopen("download.list","w");
      for(int nn=0;dll[nn];++nn){
        fprintf(ff,"%s\n",dll[nn]);
        delete[] dll[nn];
      }
      fclose(ff);
      delete[] dll;
      char cmd[1000];
      sprintf(cmd,"rsync --files-from=download.list --copy-links -prougtlvc vannoort@gofio2.mps.mpg.de:%s %s",dlpath,rawpath);
      if(system(cmd)) io.msg(IOL_ERROR,"error fetching files\n");
    }
  }
#else
  io.msg(IOL_INFO,"not checking for data\n");
#endif
  
  int ndx=-1,ndy=-1,ndl=-1;      // data dimensions
  if(cfg.get("-ndx",ndx)) io.msg(IOL_XNFO,"Setting width to %d.\n",ndx);
  if(cfg.get("-ndy",ndy)) io.msg(IOL_XNFO,"Setting height to %d.\n",ndy);
//
  int dxoffs=-1,dyoffs=-1;      // border around data coordinates
  if(cfg.get("-xb",dxoffs)) io.msg(IOL_INFO,"Border width is %d.\n",dxoffs);
  if(cfg.get("-yb",dyoffs)) io.msg(IOL_INFO,"Border height is %d.\n",dyoffs);
//
  if((ndx<0)||(ndy<0)){          // no data dimensions specified, try to read them from file!
    int *ds,dims,type,bdx=0,bdy=0,xidx=1,yidx=2;
    char *header,path[1000];
    if(readrhs){  // if we have a RHS: use that: this has the wrong dimensions (borders)!
      if(rhsname) sprintf(path,"%s",rhsname); else sprintf(path,"rhs_ff.%02d.%02d.f0",0,0);
      bdx=2*dxoffs;
      bdy=2*dyoffs;
      xidx=2;
      yidx=1;
    }else{        // no RHS: try raw data
      char *tmpname=strcpy(new char [strlen(mimg->names[0])+1],mimg->names[0]);
      char *filename=strrchr(tmpname,'/');
      filename[0]=0;
      filename+=1;
      char *basepath=strrchr(tmpname,'/');
      basepath[0]=0;
      sprintf(path,"%s/%s",rawpath,filename);
    }
//              
    io.msg(IOL_XNFO,"Reading %s\n",path);
    if(byte *data=ana_fzread(path,ds,dims,header,type)){
      ndl=ds[3];
      ndx=ds[xidx]-bdx;
      ndy=ds[yidx]-bdy;
      delete[] header;
      delete[] (ds+1);
      delete[] data;
    }else io.msg(IOL_ERROR,"Failed to acquire data dimensions!\n");
  }
  io.msg(IOL_XNFO,"Raw data dimensions: %d %d %d\n",ndx,ndy,ndl);
/*
#if 1
//NaD 20180817
  int ndx=116,ndy=128;      // data dimensions
#else
//Ha  20180824
  int ndx=114,ndy=128;      // data dimensions
#endif
*/
  int nx=ndx+dxoffs*2,ny=ndy+dyoffs*2;  // restored dimensions: 16 pixel border
  int ll=1,lh=ndl;          // wavelength range
  cfg.get("-ll",ll);
  cfg.get("-lh",lh);
  int nl=lh-ll+1;
//
  io.msg(IOL_XNFO,"Restored data dimensions: %d %d %d\n",nx,ny,lh-ll+1);
//
  int dxl=1+dxoffs,dxh=ndx+dxoffs,dyl=1+dyoffs,dyh=ndy+dyoffs; // raw data cube location in restored space

  uint08_t **mask=0;
  char *maskname=0;
  cfg.get("--mask",maskname);
  if(maskname){// read mask from file
    int *ds,dims,type;
    char *header;
    byte *data=ana_fzread(maskname,ds,dims,header,type);
    io.msg(IOL_INFO,"mask: %d %d %d %d\n",dxh-dxl+1,dyh-dyl+1,ds[2],ds[1]);
    mask=b2dim(data,dxl,dxh,dyl,dyh);
    delete[] header;
    delete[] (ds+1);
  }else{ // unmask all
    mask=b2dim(dxl,dxh,dyl,dyh);
    for(int x=dxl;x<=dxh;++x) for(int y=dyl;y<=dyh;++y) mask[x][y]=1;
  }
// auxiliary global tip/tilt correction
  fp_t tx=0.0;
  if(cfg.get("-dx",tx)) io.msg(IOL_INFO,"applying %lf pixels tilt in x\n",tx);
  fp_t ty=0.0;
  if(cfg.get("-dy",ty)) io.msg(IOL_INFO,"applying %lf pixels tilt in y\n",ty);
// initialize stretch maps
  float64_t **xpos=0;
  char *xposname=0;
  cfg.get("--xpos",xposname);
  if(xposname){// read xpos from file
    int *ds,dims,type;
    char *header;
    byte *data=ana_fzread(xposname,ds,dims,header,type);
    io.msg(IOL_INFO,"xpos: %d %d\n",ds[2],ds[1]);
    xpos=f64t2dim((float64_t*)data,1,ds[2],1,ds[1]);
    delete[] header;
    delete[] (ds+1);
  }
  float64_t **ypos=0;
  char *yposname=0;
  cfg.get("--ypos",yposname);
  if(yposname){// read ypos from file
    int *ds,dims,type;
    char *header;
    byte *data=ana_fzread(yposname,ds,dims,header,type);
    io.msg(IOL_INFO,"ypos: %d %d\n",ds[2],ds[1]);
    ypos=f64t2dim((float64_t*)data,1,ds[2],1,ds[1]);
    delete[] header;
    delete[] (ds+1);
  }
  if(xpos&&ypos) mimg->stretch(xpos,ypos,io);
//
  for(int k=0;k<=0;++k)
  for(int l=0;l<=0;++l){
// retrieve restored image
    int npts;
// save wb object
    if(imgbase){
      float64_t **img=f64t2dim(1,nx,1,ny);
      for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) img[x][y]=mimg->imgval(x-dxoffs,y-dyoffs,io);
      int ds[]={0,ny,nx};
      char name[100];
      sprintf(name,"%s.%02d.%02d.f0",imgbase,k,l);
      ana_fzwrite((byte*)(img[1]+1),name,ds,2,(char*)"bla",FLOAT64);
      del_f64t2dim(img,1,nx,1,ny);
    }
//
    io.msg(IOL_INFO,"Using %d threads\n",nthrd);
    mt_class mt(nthrd);

    int nframes=(mimg)?mimg->len():0;
    if(nframes) io.msg(IOL_INFO,"%d frames total\n",nframes);
    struct mm m(*mimg,nframes);
//
    float64_t ***rhs=0;
//
    if(!norhs){
      if(!readrhs){
        io.msg(IOL_INFO,"calculating RHS\n");
        if(rhsname&&save) io.msg(IOL_INFO,"RHS will be saved to \"%s\"\n",rhsname);
//
        rhs=f64t3dim(ll,lh,1,nx,1,ny);
        mt.mkrhs(m,npts,rhs,tx,ty,ll,lh,nx,ny,dxl,dxh,dyl,dyh,mask,mul,mod,off,rawpath,io);
        if(save){
          int ds[]={0,ny,nx,nl};
          char name[100];
          if(rhsname) sprintf(name,"%s",rhsname); else sprintf(name,"rhs_ff.%02d.%02d.f0",k,l);
          ana_fzwrite((byte*)(rhs[ll][1]+1),name,ds,3,(char*)"bla",FLOAT64);
        }
      }else{
        int *ds,dims,type;
        char *header,name[100];
        if(rhsname) sprintf(name,"%s",rhsname); else sprintf(name,"rhs_ff.%02d.%02d.f0",k,l);
        io.msg(IOL_INFO,"reading RHS from \"%s\"\n",name);
        byte *data=ana_fzread(name,ds,dims,header,type);
        if(dims==3){
          if((nl==ds[3])&&(ny==ds[1])&&(nx==ds[2])) // is this right? nx <-> ny?
            rhs=f64t3dim((float64_t*)data,ll,lh,1,nx,1,ny);
        }

        if(dims==3){
          if((ny==ds[1])&&(nx==ds[2])){ // is this right? nx <-> ny?
            if(nl==ds[3])
              rhs=f64t3dim((float64_t*)data,ll,lh,1,nx,1,ny);
            else if(lh<=ds[3]){
              io.msg(IOL_INFO,"Selecting reduced RHS wavelength range (%d-%d)\n",ll,lh);
              float64_t ***tmp=f64t3dim((float64_t*)data,1,ds[3],1,nx,1,ny);
              rhs=f64t3dim(ll,lh,1,nx,1,ny);
              for(int lm=ll;lm<=lh;++lm) memcpy(rhs[lm][1]+1,tmp[lm][1]+1,nx*ny*sizeof(float64_t));
              del_f64t3dim(tmp,1,ds[3],1,nx,1,ny);
            }else io.msg(IOL_ERROR,"RHS lambda range (%d) < requested upper wavelength (%d). Aborting!\n",ds[3],lh);
          }else io.msg(IOL_ERROR,"Dimension mismatch: RHS and set dimensions are not the same!\n");
        }else io.msg(IOL_ERROR,"Dimension mismatch: dim[RHS]=%d, expected 3!\n",dims);
        delete[] header;
        delete[] (ds+1);
      }
      if(nolhs) return 0;
    }
//
    float64_t ****A=0;
    if(!nolhs){
      if(!readmat){
        io.msg(IOL_INFO,"calculating LHS\n");
        if(lhsname&&save) io.msg(IOL_INFO,"LHS will be saved to \"%s\"\n",lhsname);
        A=f64t4dim(1,nx,1,ny,1,nx,1,ny);
        mt.mkmat(m,npts,A,nx,ny,dxl,dxh,dyl,dyh,mask,tx,ty,mul,mod,off,io);
        for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y){ // fill in the lower triangle
          for(int u=1;u<=x-1;++u) for(int v=1;v<=ny;++v)  A[x][y][u][v]=A[u][v][x][y];
          for(int v=1;v<y;++v) A[x][y][x][v]=A[x][v][x][y];
        }
        if(save){ // save A
          int ds[]={0,ny,nx,ny,nx};
          char name[100];
//          sprintf(name,"A.%02d.%02d.f0",k,l);
          if(lhsname) sprintf(name,"%s",lhsname); else sprintf(name,"A.%02d.%02d.f0",k,l);
          ana_fzwrite((byte*)(A[1][1][1]+1),name,ds,4,(char*)"bla",FLOAT64);
        }
      }else{ // read A from file
        int *ds,dims,type;
        char *header,name[100];
//        sprintf(name,"A.%02d.%02d.f0",k,l);
        if(lhsname) sprintf(name,"%s",lhsname); else sprintf(name,"A.%02d.%02d.f0",k,l);
        io.msg(IOL_INFO,"reading LHS from \"%s\"\n",name);
        byte *data=ana_fzread(name,ds,dims,header,type);
        A=f64t4dim((float64_t*)data,1,nx,1,ny,1,nx,1,ny);
        delete[] header;
        delete[] (ds+1);
      } 
    }
//
    if(noslv) return 0;
//
    float64_t cdm=0.0;
    for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) cdm+=A[x][y][x][y];
    cdm/=(float64_t)(nx*ny);
    io.msg(IOL_INFO,"mean diagonal: %5.3f\n",cdm);
//
    io.msg(IOL_INFO,"flats\n");
// correct A for flatfield response...
    float64_t **flat=f64t2dim(1,nx,1,ny);
    float64_t **fimg=f64t2dim(1,nx,1,ny);
    for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y) fimg[x][y]=1.0;
    mt.dot(m,nx,ny,A,fimg,flat,io);
    del_f64t2dim(fimg,1,nx,1,ny);
    if(save){
      int ds[]={0,ny,nx};
      char name[100];
      sprintf(name,"lhs_flat.%02d.%02d.f0",k,l);
      ana_fzwrite((byte*)(flat[1]+1),name,ds,2,(char*)"bla",FLOAT64);
    }
//
    if(!noff)
      for(int x=1;x<=nx;++x) for(int y=1;y<=ny;++y)
        for(int u=1;u<=nx;++u) for(int v=1;v<=ny;++v) A[x][y][u][v]/=flat[x][y];
    del_f64t2dim(flat,1,nx,1,ny);
//
    io.msg(IOL_INFO,"restoring\n"); // calculate RHS
//
    float64_t ***res=f64t3dim(ll,lh,1,nx,1,ny);
    memcpy(res[ll][1]+1,rhs[ll][1]+1,nl*nx*ny*sizeof(float64_t)); // start with rhs, or resume...?

      if(lr>0){
        if(lrbase) io.msg(IOL_INFO,"LR name base \"%s\"\n",lrbase);
        for(int iter=1,lrp=0;iter<=nlr;++iter){
          io.msg(IOL_INFO,"LR iterations %d - %d (%d)\n",lrp,lr[iter],lr[iter]-lrp);
          mt.solve(m,nx,ny,A,rhs,res,ll,lh,cdm,SLV_METHOD_LR,lr[iter]-lrp,io);
          int ds[]={0,ny,nx,nl};
          char name[100];
          if(lrbase) sprintf(name,"%s.%02d.%02d.lr%04d.f0",lrbase,k,l,lr[iter]); else sprintf(name,"img_fit.%02d.%02d.lr%04d.f0",k,l,lr[iter]);
          ana_fzwrite((byte*)(res[ll][1]+1),name,ds,3,(char*)"bla",FLOAT64);
          lrp=lr[iter];
        }
      }else{
        mt.solve(m,nx,ny,A,rhs,res,ll,lh,cdm,SLV_METHOD_MG,0,io);
        int ds[]={0,ny,nx,nl};
        char name[100];
        sprintf(name,"img_fit.%02d.%02d.f0",k,l);
        ana_fzwrite((byte*)(res[ll][1]+1),name,ds,3,(char*)"bla",FLOAT64);
      }

/*
    mt.solve(m,nx,ny,A,rhs,res,cdm,SLV_METHOD_LR,2000,io); 
    { // always save fitted data
      int ds[]={0,ny,nx};
      char name[100];
      sprintf(name,"img_fit.%02d.%02d.f0",k,l);
      ana_fzwrite((byte*)(res[1]+1),name,ds,2,(char*)"bla",FLOAT64);
    }
*/
    del_f64t3dim(res,ll,lh,1,nx,1,ny);   
    del_f64t4dim(A,1,nx,1,ny,1,nx,1,ny);
    del_f64t3dim(rhs,ll,lh,1,nx,1,ny);
//
  } 
  if(xpos) del_f64t2dim(xpos,1,1,1,1);
  if(ypos) del_f64t2dim(ypos,1,1,1,1);
  return 0;
}
