#ifndef __MOMFBD_READ_H__      // __MOMFBD_READ_H__
#define __MOMFBD_READ_H__

#include "types.h"
#include "io.h"
#include "momfbd_read.h"

#define MOMFBD_INT8	        0
#define MOMFBD_INT16	        1
#define MOMFBD_INT32	        2
#define MOMFBD_FLOAT32	        3
#define MOMFBD_FLOAT64	        4

struct patch_struct{
  int32_t nc;
  int32_t xl,xh,yl,yh,offx,offy;
  int32_t *dx,*dy,*nim;
  int32_t nx,ny;
  float32_t **img;
  float32_t ***psf;
  int32_t npsf;
  float32_t ***obj;
  int32_t nobj;
  float32_t ***res;
  int32_t nres;
  float32_t **alpha;
  int32_t nalpha,nm;
  float32_t ***diversity;
  int32_t ndiv,nphx,nphy;
  float64_t sdx,sdy; // stretch offsets
  float64_t **tfwf;
private:
  void read(int fd,void *buf,int sz);
//
public:
  patch_struct(int fd,uint08_t swap_endian,const char *header);
  ~patch_struct(void);
  float64_t **wavefront(int,float32_t***,float64_t&,float64_t&,float64_t);
  float64_t **diff(patch_struct*,int,int&);
  int len(void){ return nim[1]; };
  float64_t xc(void){ return sdx+0.5*(float64_t)(xl+xh); };
  float64_t yc(void){ return sdy+0.5*(float64_t)(yl+yh); };
  void stretch(float64_t sdx_in,float64_t sdy_in){ sdx=sdx_in; sdy=sdy_in; };
  float32_t **subimg(int &np_o){
    np_o=nx;
    return img;
  };
  float32_t imgval(float32_t ix,float32_t iy);
};

struct momfbd_struct{
  float32_t pix2cf,cf2pix,version;
  float32_t ***mode,**pupil;
  int32_t nm,nph,np;
  patch_struct ***patch;
  int nsx,nsy;
  int32_t nc;
  int16_t *cl_xl,*cl_xh,*cl_yl,*cl_yh;
//
  float64_t xo,yo,sf;
//
  char *filename;
  char *time_obs,*date_obs;
  char **names;
private:
  ssize_t read(int fd,void *buf,int sz);
public:
  momfbd_struct(const char*,io_class&);
  ~momfbd_struct(void);
  void map(float64_t,float64_t,float64_t,io_class &io);
  void stretch(float64_t**,float64_t**,io_class &io);
  void info(io_class &);
  float64_t **psf(int,int,int,int,float64_t&,float64_t&,int&,io_class&);
  float64_t *ds(momfbd_struct*,int,int,io_class&);
  float64_t *ds(momfbd_struct*,int,int,int,io_class&);
  int len(void){ return patch[1][1]->len(); };
  int centerpatch(int,int);
  void centerpatch(int xc,int yc,int &sx,int &sy);
  float64_t **subimg(int&,int,int,io_class&);
  char **get_names(void){ return names; };
  char **fetchdata(char**,int,char*,io_class &io);
  char *fetchdatapath(io_class &io);
//  void map(int,int,float64_t,io_class &io);
  float64_t imgval(int ix_in,int iy_in,io_class &io);
};

//byte *momfbd_read(char *file_name,int *&ds,int &nd,char *&header,int &type,io_class &io);

#endif                         // __MOMFBD_READ_H__
