#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "types.h"
#include "anadecompress.h"

#include "uts.h"
#include "ana_io.h"

int ck_synch_hd(FILE *fin,struct fzhead &fh,int t_endian)
{
  int wwflag=0;
  if(fread(&fh,1,sizeof(struct fzhead),fin)!=sizeof(struct fzhead)) exit(fprintf(stderr,"error in fzread while reading header\n"));
  int syncpat=(fh.synch_pattern==0x5555aaaa);
  int revsyncpat=(fh.synch_pattern==0xaaaa5555);
  if(!(syncpat||revsyncpat)){
    fclose(fin);
    exit(fprintf(stderr,"error: file does not have the F0 synch pattern (found 0x%x instead)\n",fh.synch_pattern));
  }
  if(syncpat==t_endian){
    fprintf(stdout,"warning: reversed F0 synch pattern\n");
    wwflag=1; 
  }
  if(fh.nhb>1){                     // if the header is long, read in the rest now
    if(fh.nhb>15){
      fclose(fin);
      exit(fprintf(stderr,"error: cannot handle header more than 16 blocks!\n"));
    }
    int size=(fh.nhb-1)*sizeof(struct fzhead);
    byte *buf=new byte [size];
    if(fread(buf,1,size,fin)<size) exit(fprintf(stderr,"error: %s!\n",strerror(errno)));
    delete buf; // not very useful?
  }
  if(t_endian) swap(fh.dim,fh.ndim); // for big endian machines
  return wwflag; 
}

byte *ana_fzread(char *file_name,int *&ds,int &nd,char *&header,int &type) // fzread subroutine	
{
  struct stat stat_buf;
  if(stat(file_name,&stat_buf)<0){
//    fprintf(stdout,"file \"%s\" not found.\n",file_name);  
    return 0;
  }
//  fprintf(stdout,"reading ANA file \"%s\"\n",file_name);
  int type_sizes[]=ANA_VAR_SZ;
  int one=1;
  int t_endian=(*(char*)&one==0);                      // an endian detector, taken from SL's tiff library 
//
  FILE *fin=fopen(file_name,"r");
  if(!fin){
//    fprintf(stdout,"file \"%s\" not found!\n",file_name);
    return 0;
  }
  struct fzhead	fh;
  int sef;
  if((sef=ck_synch_hd(fin,fh,t_endian))<0) return 0;
  header=strcpy(new char [strlen(fh.txt)+1],fh.txt);
  ds=new int [nd=fh.ndim]-1;
  for(int d=1;d<=nd;++d) ds[d]=fh.dim[d-1];

  int n_elem=1;
  for(int d=0;d<=fh.ndim-1;++d) n_elem*=fh.dim[d]; // compute size of array
  type=fh.datyp;
  int f_endian=(fh.subf>=128);                     // the top bit of the byte fh->subf denotes endian type, 0 for little and 1 for big
  int swap_endian=(f_endian!=t_endian);            // file has different endianness
  if(sef) swap_endian=(!swap_endian);              // file contains strange data
  int compressed=(fh.subf&1);                      // file is in compressed format
  if(compressed){                                  // compressed format
    struct compresshead ch;
    if(fread(&ch,1,14,fin)<14) exit(fprintf(stderr,"error reading in compression header\n"));
// header by default little-endian?
    if(t_endian){ // big endian platform
      swap(&ch.tsize,1);
      swap(&ch.nblocks,1);
      swap(&ch.bsize,1);
    }
// read data
    int size=ch.tsize-14;
    byte *buf=new byte [size+4];
    if(fread(buf,1,size,fin)<size) exit(fprintf(stderr,"error reading in compressed data\n"));
    fclose(fin);
//
    if(ch.bsize*ch.nblocks>n_elem){ // fix a possible problem with ch.nblocks
      fprintf(stdout,"warning, bad ch.nblocks = %d\ncorrecting to %d, hope this is right!\n",ch.nblocks,n_elem/ch.bsize);
      ch.nblocks=n_elem/ch.bsize;
    }
    if(ch.type%2==type) exit(fprintf(stderr,"inconsistent compression type\n")); // consistency check
    int rv;
    byte *out=new byte [n_elem*type_sizes[type]];
    switch(ch.type){
      case(0): rv=anadecrunch(buf,(int16_t*)out,ch.slice_size,ch.bsize,ch.nblocks,t_endian==MFBD_LITTLE_ENDIAN); break;
      case(1): rv=anadecrunch8(buf,(int08_t*)out,ch.slice_size,ch.bsize,ch.nblocks,t_endian==MFBD_LITTLE_ENDIAN); break;
      case(2): rv=anadecrunchrun(buf,(int16_t*)out,ch.slice_size,ch.bsize,ch.nblocks,t_endian==MFBD_LITTLE_ENDIAN); break;
      case(3): rv=anadecrunchrun8(buf,(int08_t*)out,ch.slice_size,ch.bsize,ch.nblocks,t_endian==MFBD_LITTLE_ENDIAN); break;
      case(4): rv=anadecrunch32(buf,(int32_t*)out,ch.slice_size,ch.bsize,ch.nblocks,t_endian==MFBD_LITTLE_ENDIAN); break;
      default: exit(fprintf(stderr,"error in data type for compressed data, fh.datyp =%d\n",fh.datyp));
    }
    delete[] buf;
    return out;
  }else{                            // uncompressed
    size_t size=(size_t)n_elem*(size_t)type_sizes[type];
    byte *out=new byte [size];
    if(fread(out,1,size,fin)<size){
      fclose(fin);
      exit(fprintf(stderr,"error: unexpected end of file\n"));
    }
    fclose(fin);
    if(swap_endian) // endianness is wrong
      switch(type){
        case(INT16): swap((int16_t*)out,n_elem); break;
        case(INT32):
        case(FLOAT32): swap((int32_t*)out,n_elem); break;
        case(FLOAT64): swap((int64_t*)out,n_elem); break;
      }
    return out; 
  } // end if(compressed)
}
