#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "types.h"
#include "uts.h"
#include "anacompress.h"
#include "ana_io.h"

void ana_fcwrite(byte *data,char *file_name,int *ds,int nd,char *header,int type)	/* fcwrite subroutine */
{ // write standard f0 files, compressed format
  FILE *f=fopen(file_name,"w");
  struct fzhead fh;
  memset(&fh,0,sizeof(struct fzhead));
  int one=1;
  int t_endian=(*(char*)&one==0);    // an endian detector, taken from SL's tiff library 
  if(t_endian){ // BIG_ENDIAN
    fh.synch_pattern=0xaaaa5555;
    fh.subf=129;
  }else{        // LITTLE_ENDIAN
    fh.synch_pattern=0x5555aaaa;
    fh.subf=1;
  }
  fh.source=0;
  fh.nhb=1;					// may be changed later
  fh.datyp=type;
  fh.ndim=nd;
//
  int n_elem=1;
  for(int n=1;n<=nd;++n) n_elem*=(fh.dim[n-1]=ds[n]);
  int nx=fh.dim[0];
  int ny=n_elem/nx;
  int type_sizes[]=ANA_VAR_SZ;
  int size=n_elem*type_sizes[type];
  if(t_endian){ // big endian platform
    switch(type){
      case(INT16): swap((int16_t*)data,n_elem); break;
      case(INT32):
      case(FLOAT32): swap((int32_t*)data,n_elem); break;
      case(FLOAT64): swap((int64_t*)data,n_elem); break;
    }
    swap(fh.dim,nd);
  }
  if(header){
    int len=min(strlen(header),255);
    strncpy(fh.txt,header,len);
    fh.txt[len]=0;
  }
/* now compress the array, must be a byte or short */
/* extended to 32 bits 2/4/96 */
  int limit=size+size/2,res,crunch_slice=5,runlengthflag=0;
  byte *q=new byte [size+size/2]; // a bit more, just in case
  switch(type){
    case(INT8):{
      if(runlengthflag)
        res=anacrunchrun8(q,data,crunch_slice,nx,ny,limit,t_endian);
      else
        res=anacrunch8(q,data,crunch_slice,nx,ny,limit,t_endian);
      break;
    }
    case(INT16):{
      if(runlengthflag)
        res=anacrunchrun(q,(int16_t*)data,crunch_slice,nx,ny,limit,t_endian);
      else
        res=anacrunch(q,(int16_t*)data,crunch_slice,nx,ny,limit,t_endian);
      break;
    }
    case(INT32):{
      if(runlengthflag){
        fprintf(stdout,"FCRUNWRITE not supported for I*4 yet\n");
        fclose(f);
        delete q;
        return;
      }else
        res=anacrunch32(q,(int32_t*)data,crunch_slice,nx,ny,limit,t_endian);
      break;
    }
    default:{
      fprintf(stdout,"FCWRITE: unsupported type\n");
      fclose(f);
      delete q;
      return;
    }
  }
  if(res<0){
    fprintf(stdout,"not enough space allocated (%d bytes) for compressed array\n",limit);
    fclose(f);
    delete q;
    return;
  }
  size=res;
  if(t_endian) swap(&res,1);
  for(int i=0;i<=3;++i) fh.cbytes[i]=((byte*)&res)[i];
  fwrite(&fh,1,sizeof(struct fzhead),f);  // write header
  fwrite(q,1,size,f);
  fclose(f);
  delete q;
}

void ana_fzwrite(byte *data,char *file_name,int *ds,int nd,char *header,int type)	/* fcwrite subroutine */
{ // write standard f0 files, compressed format
  FILE *f=fopen(file_name,"w");
  struct fzhead fh;
  memset(&fh,0,sizeof(struct fzhead));
  int one=1;
  int t_endian=(*(char*)&one==0);    // an endian detector, taken from SL's tiff library 
  if(t_endian){ // BIG_ENDIAN
    fh.synch_pattern=0xaaaa5555;
  }else{        // LITTLE_ENDIAN
    fh.synch_pattern=0x5555aaaa;
  }
  fh.source=0;
  fh.nhb=1;					// may be changed later
  fh.datyp=type;
  fh.ndim=nd;
//
  int n_elem=1;
  for(int n=1;n<=nd;++n) n_elem*=(fh.dim[n-1]=ds[n]);
  int nx=fh.dim[0];
  int ny=n_elem/nx;
  int type_sizes[]=ANA_VAR_SZ;
  size_t size=(size_t)n_elem*(size_t)type_sizes[type];
  if(t_endian){ // big endian platform
    switch(type){
      case(INT16): swap((int16_t*)data,n_elem); break;
      case(INT32):
      case(FLOAT32): swap((int32_t*)data,n_elem); break;
      case(FLOAT64): swap((int64_t*)data,n_elem); break;
    }
    swap(fh.dim,nd);
  }
  if(header){
    int len=min(strlen(header),255);
    strncpy(fh.txt,header,len);
    fh.txt[len]=0;
  }
  int res=size;
  if(t_endian) swap(&res,1);
  for(int i=0;i<=3;++i) fh.cbytes[i]=((byte*)&res)[i];
  fwrite(&fh,sizeof(struct fzhead),1,f);  // write header
  fwrite(data,1,size,f);
  fclose(f);
}
