#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <pthread.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "types.h"
#include "uts.h"
#include "mem.h"
#include "ana_io.h"
//#include "fits_read.h"
#include "momfbd_read.h"
#include "cmdcfg.h"

char **append(char **a,const char *b)
{
  int la=0;
  if(a) while(a[la]) ++la;
//
  char **tmp=new char* [la+2];
//
  if(la) memcpy(tmp,a,la*sizeof(char*));
  tmp[la]=strcpy(new char [strlen(b)+1],b);
  tmp[la+1]=0;
//
  return tmp;
}

int main(int argc,char *argv[])
{ 
  const char *options[]=CMDOPTS,*maps[]=CMDMAPS;    // recognised options
  cmdln cfg(argc,argv,options,maps);
  io_class io(cfg);
//
  int kl=1;
  cfg.get("-kl",kl);
//
  int kh=9;
  cfg.get("-kh",kh);
/*
  int ll=1;
  cfg.get("-ll",ll);
//
  int lh=1934;
  cfg.get("-lh",lh);
*/
  int nthrd=1;
  cfg.get("-n",nthrd);
//
  int readrhs=0;
  cfg.get("-r",readrhs);
//
  int readmat=0;
  cfg.get("-a",readmat);
//
  int readres=0;
  cfg.get("-x",readres);
//
  int save=0;
  cfg.get("-s",save);
//
  int noff=0;
  cfg.get("--noff",noff);
  int nolhs=0;
  cfg.get("--nolhs",nolhs);
  int norhs=0;
  cfg.get("--norhs",norhs);
  int noslv=nolhs;
  if(!noslv) cfg.get("--noslv",noslv);
//
  int po=0;
  cfg.get("--poloffs",po);
  int ps=0;
  cfg.get("--polskip",ps);
//
  int *lr=0,nlr=-1;
  char *lrstr=0;
  cfg.get("-lr",lrstr);
  if(lrstr){
    get_numbers(lrstr,lr,nlr);
    if(!nlr){
      io.msg(IOL_ERROR,"LR numbers %s did not expand to any numbers\n",lrstr);
    }else{
      char *p=new char [nlr*8];
      p[0]=0;
      for(int j=1;j<=nlr;++j) sprintf(p+strlen(p)," %d",lr[j]);
      io.msg(IOL_INFO,"LR: using %s iterations\n",p);
      delete[] p;
    }
    delete[] lrstr;
  }
//
  int di=0;
  if(cfg.get("-di",di))
    io.msg(IOL_DFLT,"Restored frame separation set to %d exposures.\n",di);
   else
    io.msg(IOL_ERROR,"Restored frame separation not specified (hint: use -di <number>).\n",di);
//
// =========================================
//
  int nd=0;
  char **list=0;
  FILE *f=fopen(argv[1],"r");
  if(!f) io.msg(IOL_ERROR,"failed to open %s\n",argv[1]);
  int n=0;
  list=new char* [1] -1;
  list[1]=0;
  char s[1000];
  while(fgets(s,1000,f)){
    char **tmp=new char* [n+2] -1;
    memcpy(tmp+1,list+1,n*sizeof(char*));
    delete[] (list+1);
    list=tmp;
    list[++n]=strcpy(new char [strlen(s)+1],s);
    list[n+1]=0;
    while(char *p=strchr(list[n],'\n')) p[0]=0;
  }
  nd=n;
  fclose(f);
  for(int nn=1;nn<=nd;++nn) fprintf(stderr,"%s\n",list[nn]);
//
  char *path=strcpy(new char [strlen(argv[1])+1],argv[1]);
  if(char *p=strrchr(path,'/')) p[1]=0; else sprintf(path,"./");
// remote operation
  char *momfbdpath=strcpy(new char [strlen(list[1])+1],list[1]);
  if(char *q=strrchr(momfbdpath,'/')) q[0]=0;
//
  char **dll=new char* [1];
  dll[0]=0;
  for(int nn=1;nn<=nd;++nn){
    struct stat buf;
    if(::stat(list[nn],&buf)){ // file not found?
      char cmd[1000];
      if(char *q=strrchr(list[nn],'/')){
        char *newname=new char[strlen(q)+strlen("momfbd/")+1];
        sprintf(newname,"momfbd/%s",q+1);
        delete[] list[nn];
        list[nn]=newname;
        sprintf(cmd,"%s",q+1);
      }
      if(::stat(list[nn],&buf)) dll=append(dll,cmd); // still not found: list for download
    }
  }
//
  if(dll[0]){
    FILE *ff=fopen("download.list","w");
    for(int nn=0;dll[nn];++nn){
      fprintf(ff,"%s\n",dll[nn]);
      delete[] dll[nn];
    }
    fclose(ff);
    delete[] dll;
    io.msg(IOL_ERROR,"error fetching files!\n");
  }
//
  float64_t sf=0.830769;   // x_momfbd=(x_restored-xoffs)/sf
  float64_t xoffs=419.0;   // x_restored=0 -> x_momfbd=xoffs
  float64_t yoffs=440.0;   // y_restored=0 -> y_momfbd=yoffs
//
  int ndx=114,ndy=128;      // data dimensions
  int nx=ndx+32,ny=ndy+32;  // restored dimensions: 16 pixel border
//
  momfbd_struct* mp=new momfbd_struct(list[1],io);
//  mp->map(-xoffs*sf,-yoffs*sf,sf,io);
  int nsx=mp->nsx,nsy=mp->nsy;

  io.msg(IOL_INFO,"nx=%d ny=%d\n",nsx,nsy);

  float64_t ***xpos=f64t3dim(1,nd,1,nsx,1,nsy);
  float64_t ***ypos=f64t3dim(1,nd,1,nsx,1,nsy);
  memset(xpos[1][1]+1,0,nd*nsx*nsy*sizeof(float64_t));
  memset(ypos[1][1]+1,0,nd*nsx*nsy*sizeof(float64_t));
  
  float64_t **xpm=f64t2dim(1,nsx,1,nsy),**ypm=f64t2dim(1,nsx,1,nsy);
  memset(xpm[1]+1,0,nsx*nsy*sizeof(float64_t));
  memset(ypm[1]+1,0,nsx*nsy*sizeof(float64_t));
  for(int nn=2;nn<=nd;++nn){
    fprintf(stderr,"%5.1f%%\r",100.0*(float)(nn-1)/(float)(nd-1));
    momfbd_struct* mc=new momfbd_struct(list[nn],io);
//    mc->map(-xoffs*sf,-yoffs*sf,sf,io);
    for(int x=1;x<=nsx;++x)
      for(int y=1;y<=nsy;++y){
        float64_t *ds=mc->ds(mp,di,x,y,io);
        if(isfinite(ds[1])&&isfinite(ds[2])&&(fabs(ds[1])<10.0)&&(fabs(ds[2])<10.0)){
          xpm[x][y]+=(xpos[nn][x][y]=xpos[nn-1][x][y]+ds[1]);
          ypm[x][y]+=(ypos[nn][x][y]=ypos[nn-1][x][y]+ds[2]);
        }else{
          io.msg(IOL_WARN,"bad displacement between %s and %s in patch [%d,%d]: %f %f.\n",list[nn-1],list[nn],x,y,ds[1],ds[2]);
        }
        delete[] (ds+1);
      }
    delete mp;
    mp=mc;
  }
// normalize mean position... 
  for(int x=1;x<=nsx;++x)
    for(int y=1;y<=nsy;++y){
      xpm[x][y]/=(float64_t)nd;
      ypm[x][y]/=(float64_t)nd;
    }
// subtract mean position... 
  for(int nn=1;nn<=nd;++nn)
    for(int x=1;x<=nsx;++x)
      for(int y=1;y<=nsy;++y){
        xpos[nn][x][y]-=xpm[x][y];
        ypos[nn][x][y]-=ypm[x][y];
      }
  del_f64t2dim(xpm,1,nsx,1,nsy);
  del_f64t2dim(ypm,1,nsx,1,nsy);
  int aw=0;
  if(cfg.get("-aw",aw)){
    io.msg(IOL_DFLT,"Removing running average over %d frames.\n",aw);
// calculate smoothed positions
    float64_t ***sxpos=f64t3dim(1,nd,1,nsx,1,nsy);
    float64_t ***sypos=f64t3dim(1,nd,1,nsx,1,nsy);
    float64_t **nnpos=f64t2dim(1,nsx,1,nsy);
    memset(sxpos[1][1]+1,0,nd*nsx*nsy*sizeof(float64_t));
    memset(sypos[1][1]+1,0,nd*nsx*nsy*sizeof(float64_t));
//
    for(int nn=1;nn<=nd;++nn){
      int ml=max(nn-aw/2,1);
      int mh=min(nn+aw/2,nd);
//      ml=mh-aw+1;
      memset(nnpos[1]+1,0,nsx*nsy*sizeof(float64_t));
      for(int mm=ml;mm<=mh;++mm){
        for(int x=1;x<=nsx;++x)
          for(int y=1;y<=nsy;++y){
            sxpos[nn][x][y]+=xpos[mm][x][y];
            sypos[nn][x][y]+=ypos[mm][x][y];
            nnpos[x][y]+=1.0;
          }
      }
      for(int x=1;x<=nsx;++x)
        for(int y=1;y<=nsy;++y){
          sxpos[nn][x][y]/=nnpos[x][y];
          sypos[nn][x][y]/=nnpos[x][y];
        }
    }
// subtract smoothed positions?
    for(int nn=1;nn<=nd;++nn)
      for(int x=1;x<=nsx;++x)
        for(int y=1;y<=nsy;++y){
          xpos[nn][x][y]-=sxpos[nn][x][y];
          ypos[nn][x][y]-=sypos[nn][x][y];
        }
    del_f64t2dim(nnpos,1,nsx,1,nsy);
    {
      int ds[]={0,nsy,nsx,nd};
      char name[100];
      sprintf(name,"sxpos.f0");
      ana_fzwrite((byte*)(sxpos[1][1]+1),name,ds,3,(char*)"bla",FLOAT64);
      sprintf(name,"sypos.f0");
      ana_fzwrite((byte*)(sypos[1][1]+1),name,ds,3,(char*)"bla",FLOAT64);
    }
//
    del_f64t3dim(sxpos,1,nd,1,nsx,1,nsy);
    del_f64t3dim(sypos,1,nd,1,nsx,1,nsy);
  }else io.msg(IOL_DFLT,"Running average frames not specified, using raw displacements!.\n");
 
  {
    int ds[]={0,nsy,nsx,nd};
    char name[100];
    sprintf(name,"xpos.f0");
    ana_fzwrite((byte*)(xpos[1][1]+1),name,ds,3,(char*)"bla",FLOAT64);
    sprintf(name,"ypos.f0");
    ana_fzwrite((byte*)(ypos[1][1]+1),name,ds,3,(char*)"bla",FLOAT64);
  }
  
  del_f64t3dim(xpos,1,nd,1,nsx,1,nsy);
  del_f64t3dim(ypos,1,nd,1,nsx,1,nsy);

/*
  io.msg(IOL_INFO,"reading xpos and ypos\n");
  {
    int *ds,dims,type;
    char *header,name[100];
    byte *data=ana_fzread(xpos_name,ds,dims,header,type);
    if(ds[2]!=nd) io.msg(IOL_ERROR,"xpos file %s contains %d restored files, but config file %s contains %d!\n",xpos_name,ds[2],argv[1],nd);
    xpos=f64t2dim((float64_t*)data,1,ds[2],1,nslit=ds[1]);
    delete[] header;
    delete[] (ds+1);
  }
  {
    int *ds,dims,type;
    char *header,name[100];
    byte *data=ana_fzread(ypos_name,ds,dims,header,type);
    if(ds[2]!=nd) io.msg(IOL_ERROR,"ypos file %s contains %d restored files, but config file %s contains %d!\n",ypos_name,ds[2],argv[1],nd);
    if(ds[1]!=nslit) io.msg(IOL_ERROR,"ypos slit length is %d, but xpos slit length was %d!\n",ds[2],nslit);
    ypos=f64t2dim((float64_t*)data,1,ds[2],1,ds[1]);
    delete[] header;
    delete[] (ds+1);
  }
*/
  return 0;
}
